import os
import numpy as np

from pystog.utils import RealSpaceHeaders, ReciprocalSpaceHeaders
from pystog.converter import Converter
from pystog.transformer import Transformer

TEST_DIR = os.path.dirname(os.path.abspath(__file__))
TEST_DATA_DIR = os.path.join(TEST_DIR, "test_data")
if not os.path.exists(TEST_DATA_DIR):
    raise RuntimeError('Failed to find "test_data" directory')
TUTORIAL_DATA_DIR = os.path.join(TEST_DIR, "../tutorials/data")

# ------------------------------------------------
# General utility functions


def multiplied_template(num):
    return ('%f ' * len(num))[:-1] % tuple(num)

# Test data utility functions


def load_lammps_rdf(filename):
    test_file_path = os.path.join(TEST_DATA_DIR, filename)
    i, r, gr, nr = np.loadtxt(test_file_path, unpack=True, skiprows=5)
    return r, gr


def get_data_path(filename):
    return os.path.join(TEST_DATA_DIR, filename)


def load_data(filename, skiprows=2):
    test_file_path = get_data_path(filename)
    data = np.loadtxt(test_file_path, skiprows=skiprows)
    return data


def get_index_of_function(func_string, headers):
    for i, func_type in enumerate(headers):
        if func_type == func_string:
            return i
    return None


def write_functions_to_file(headers, data, output_filename, datadir='.'):
    header_string = ' '.join(headers)
    outfile_path = os.path.join(datadir, output_filename)
    with open(outfile_path, "w") as f:
        f.write("%d\n" % len(data))
        f.write("# %s\n" % header_string)
        for row in data:
            f.write("%s\n" % multiplied_template(row))

# ------------------------------------------------
# Real-space test data utility functions


def create_real_space_functions(gr_filename, **kwargs):
    r, gr = load_lammps_rdf(gr_filename)
    c = Converter()
    GofR = c.g_to_G(r, gr, **kwargs)
    GKofR = c.g_to_GK(r, gr, **kwargs)
    data = [r, gr, GofR, GKofR]
    assert len(RealSpaceHeaders) == len(data)
    data = np.transpose(data)  # puts the data where each column is a function
    return RealSpaceHeaders, data


def create_and_write_real_space_functions(
        gr_filename, output_filename, **kwargs):
    headers, data = create_real_space_functions(gr_filename, **kwargs)
    write_functions_to_file(
        headers,
        data,
        output_filename,
        datadir=TEST_DATA_DIR)
    write_functions_to_file(
        headers,
        data,
        output_filename,
        datadir=TUTORIAL_DATA_DIR)

# ------------------------------------------------
# Real-space test data utility functions


def create_reciprocal_space_functions(
        gr_filename,
        dq=0.02,
        qmin=0.00,
        qmax=35.0,
        **kwargs):
    r, gr = load_lammps_rdf(gr_filename)
    q = np.arange(qmin, qmax + dq, dq)
    t = Transformer()
    q, sq = t.g_to_S(r, gr, q, **kwargs)
    q, fq = t.g_to_F(r, gr, q, **kwargs)
    q, fq_keen = t.g_to_FK(r, gr, q, **kwargs)
    q, dcs = t.g_to_DCS(r, gr, q, **kwargs)
    data = [q, sq, fq, fq_keen, dcs]
    assert len(ReciprocalSpaceHeaders) == len(data)
    data = np.transpose(data)  # puts the data where each column is a function
    return ReciprocalSpaceHeaders, data


def create_and_write_reciprocal_space_functions(
        gr_filename, output_filename, **kwargs):
    headers, data = create_reciprocal_space_functions(gr_filename, **kwargs)
    write_functions_to_file(
        headers,
        data,
        output_filename,
        datadir=TEST_DATA_DIR)
    write_functions_to_file(
        headers,
        data,
        output_filename,
        datadir=TUTORIAL_DATA_DIR)

# ------------------------------------------------
# Real and Reciprocal test data utility function


def create_and_write_both_type_of_functions(
        gr_filename, real_out, reciprocal_out, **kwargs):
    headers, data = create_real_space_functions(gr_filename, **kwargs)
    write_functions_to_file(headers, data, real_out, datadir=TEST_DATA_DIR)
    write_functions_to_file(headers, data, real_out, datadir=TUTORIAL_DATA_DIR)
    headers, data = create_reciprocal_space_functions(gr_filename, **kwargs)
    write_functions_to_file(
        headers,
        data,
        reciprocal_out,
        datadir=TEST_DATA_DIR)
    write_functions_to_file(
        headers,
        data,
        reciprocal_out,
        datadir=TUTORIAL_DATA_DIR)



# -------------------------------------#
# Utilities

from collections import OrderedDict
import numpy as np

ReciprocalSpaceChoices = OrderedDict([("S(Q)", "S(Q)"),
                                      ("Q[S(Q)-1]", "also known as F(Q)"),
                                      ("FK(Q)", "Keen's F(Q)"),
                                      ("DCS(Q)", "Differential Cross-Section")])
RealSpaceChoices = OrderedDict([("g(r)", ' "little" g(r)'),
                                ("G(r)", "Pair Distribution Function"),
                                ("GK(r)", "Keen's G(r)")])

RealSpaceHeaders = ['r'] + list(RealSpaceChoices.keys())
ReciprocalSpaceHeaders = ['Q'] + list(ReciprocalSpaceChoices.keys())


def create_domain(xmin, xmax, xdelta):
    return np.arange(xmin, xmax + xdelta, xdelta)



"""
==============
PyStoG CLI
==============

This module defines the PyStoG CLI function that performs the workflow of the
original FORTRAN-based StoG CLI
"""


import json

from pystog import StoG
from pystog.stog import NoInputFilesException
from pystog.io import get_cli_parser, parse_cli_args


def pystog_cli(kwargs=None):
    """
    Main entry point for PyStoG CLI tool

    :param kwargs: Dict with CLI arguments.
                   If None, parsed from command line via get_cli_parser
    :type kwargs: dict
    """
    parser = get_cli_parser()

    if not kwargs:
        args = parser.parse_args()
        if args.json:
            print("loading config from '%s'" % args.json)
            with open(args.json, 'r') as f:
                kwargs = json.load(f)

        else:
            kwargs = parse_cli_args(args)

    if not kwargs.get("Files"):
        parser.print_help()
        raise NoInputFilesException("No input files given in arguments")

    # Merge S(Q) files
    stog = StoG(**kwargs)
    stog.read_all_data(skiprows=3)
    stog.merge_data()
    stog.write_out_merged_sq()

    # Initial S(Q) -> g(r) transform
    stog.transform_merged()
    stog.write_out_merged_gr()

    # TODO: Add the lowR minimizer here
    # print stog.get_lowR_mean_square()

    # Set the S(Q) and g(r) if no Fourier Filter
    r = stog.r_master[stog.gr_title]
    q = stog.q_master[stog.sq_title]
    sq = stog.sq_master[stog.sq_title]
    gr_out = stog.gr_master[stog.gr_title]

    # Apply Fourier Filter
    if "FourierFilter" in kwargs:
        q, sq, r, gr_out = stog.fourier_filter()

    # Apply Lorch Modification
    if kwargs["LorchFlag"]:
        r, gr_out = stog.apply_lorch(q, sq, r)

    # Apply final scale number
    stog._add_keen_fq(q, sq)
    stog._add_keen_gr(r, gr_out)


if __name__ == "__main__":
    pystog_cli()



import numpy as np
import os
import tempfile
import unittest

from tests.utils import \
    get_data_path, \
    get_index_of_function, \
    load_data
from tests.materials import Argon
from pystog.utils import \
    RealSpaceHeaders, \
    ReciprocalSpaceHeaders
from pystog.stog import NoInputFilesException, StoG


class TestStogBase(unittest.TestCase):
    rtol = 1.0
    atol = 1.0

    def initialize_material(self):
        # setup input data
        self.kwargs = self.material.kwargs

        # setup the tolerance
        self.first = self.material.reciprocal_space_first
        self.last = self.material.reciprocal_space_last

        data = load_data(self.material.reciprocal_space_filename)
        self.q = data[:, get_index_of_function("Q", ReciprocalSpaceHeaders)]
        self.sq = data[:, get_index_of_function(
            "S(Q)", ReciprocalSpaceHeaders)]
        self.fq = data[:, get_index_of_function(
            "Q[S(Q)-1]", ReciprocalSpaceHeaders)]
        self.fq_keen = data[:, get_index_of_function(
            "FK(Q)", ReciprocalSpaceHeaders)]
        self.dcs = data[:, get_index_of_function(
            "DCS(Q)", ReciprocalSpaceHeaders)]

        # targets for 1st peaks
        self.sq_target = self.material.sq_target
        self.dsq_target = 3.087955
        self.fq_target = self.material.fq_target
        self.fq_keen_target = self.material.fq_keen_target
        self.dcs_target = self.material.dcs_target

        # setup the first, last indices
        self.real_space_first = self.material.real_space_first
        self.real_space_last = self.material.real_space_last

        data = load_data(self.material.real_space_filename)
        self.r = data[:, get_index_of_function("r", RealSpaceHeaders)]
        self.gofr = data[:, get_index_of_function("g(r)", RealSpaceHeaders)]
        self.GofR = data[:, get_index_of_function("G(r)", RealSpaceHeaders)]
        self.GKofR = data[:, get_index_of_function("GK(r)", RealSpaceHeaders)]

        self.gofr_target = self.material.gofr_target
        self.GofR_target = self.material.GofR_target
        self.GKofR_target = self.material.GKofR_target

        self.gofr_ff_target = self.material.gofr_ff_target
        self.GofR_ff_target = self.material.GofR_ff_target
        self.GKofR_ff_target = self.material.GKofR_ff_target

        self.gofr_lorch_target = self.material.gofr_lorch_target
        self.GofR_lorch_target = self.material.GofR_lorch_target
        self.GKofR_lorch_target = self.material.GKofR_lorch_target

    def setUp(self):
        unittest.TestCase.setUp(self)
        self.material = Argon()
        self.initialize_material()

        self.real_xtarget = 3.525
        self.reciprocal_xtarget = 1.94
        self.fourier_filter_cutoff = 1.5

        filename = get_data_path(self.material.reciprocal_space_filename)
        self.kwargs_for_files = {
            'Files': [
                {
                    'Filename': filename,
                    'ReciprocalFunction': 'S(Q)',
                    'Qmin': 0.02,
                    'Qmax': 15.0,
                    'Y': {
                        'Offset': 0.0,
                        'Scale': 1.0
                    },
                    'X': {
                        'Offset': 0.0
                    }
                },
                {
                    'Filename': filename,
                    'ReciprocalFunction': 'S(Q)',
                    'Qmin': 1.90,
                    'Qmax': 35.2,
                    'Y': {
                        'Offset': 0.0,
                        'Scale': 1.0},
                    'X': {
                        'Offset': 0.0}
                }
            ]
        }

        self.kwargs_for_stog_input = {
            'NumberDensity': self.material.kwargs['rho'],
            '<b_coh>^2': self.material.kwargs['<b_coh>^2'],
            '<b_tot^2>': self.material.kwargs['<b_tot^2>'],
            'FourierFilter': {'Cutoff': self.fourier_filter_cutoff},
            'OmittedXrangeCorrection': False,
            'Rdelta': self.r[1] - self.r[0],
            'Rmin': min(self.r),
            'Rmax': max(self.r)
        }

    def tearDown(self):
        unittest.TestCase.tearDown(self)


class TestStogInit(TestStogBase):
    def setUp(self):
        super(TestStogInit, self).setUp()

    def test_stog_init(self):
        stog = StoG()
        self.assertEqual(stog.xmin, 100)
        self.assertEqual(stog.xmax, 0)
        self.assertEqual(stog.qmin, None)
        self.assertEqual(stog.qmax, None)
        self.assertEqual(stog.files, None)
        self.assertEqual(stog.sq_title, "S(Q) Merged")
        self.assertEqual(stog.qsq_minus_one_title, "Q[S(Q)-1] Merged")
        self.assertEqual(stog.sq_ft_title, "S(Q) FT")
        self.assertEqual(stog.fq_title, "F(Q) Merged")
        self.assertEqual(stog.real_space_function, "g(r)")
        self.assertEqual(stog.gr_title, "g(r) Merged")
        self.assertEqual(stog.gr_ft_title, "g(r) FT")
        self.assertEqual(stog.gr_lorch_title, "g(r) FT Lorched")
        self.assertEqual(stog.GKofR_title, "G(r) (Keen Version)")
        self.assertEqual(stog.rmin, 0.0)
        self.assertEqual(stog.rmax, 50.0)
        self.assertEqual(stog.rdelta, 0.01)
        self.assertEqual(stog.density, 1.0)
        self.assertEqual(stog.bcoh_sqrd, 1.0)
        self.assertEqual(stog.btot_sqrd, 1.0)
        self.assertEqual(stog.low_q_correction, False)
        self.assertEqual(stog.lorch_flag, False)
        self.assertEqual(stog.fourier_filter_cutoff, None)
        self.assertEqual(
            stog.merged_opts, {
                "Y": {
                    "Offset": 0.0, "Scale": 1.0}})
        self.assertEqual(stog.stem_name, "out")
        self.assertEqual(stog.reciprocal_individuals.size, 0)
        self.assertEqual(stog.q_master, {})
        self.assertEqual(stog.sq_master, {})
        self.assertEqual(stog.sq_individuals.size, 0)
        self.assertEqual(stog.r_master, {})
        self.assertEqual(stog.gr_master, {})

    def test_stog_init_kwargs_files(self):
        stog = StoG(**{'Files': ['file1.txt', 'file2.txt']})
        self.assertEqual(stog.files, ['file1.txt', 'file2.txt'])

    def test_stog_init_kwargs_real_space_function(self):
        stog = StoG(**{'RealSpaceFunction': 'G(r)'})
        self.assertEqual(stog.real_space_function, 'G(r)')

    def test_stog_init_kwargs_rmin(self):
        stog = StoG(**{'Rmin': 2.0})
        self.assertEqual(stog.rmin, 2.0)

    def test_stog_init_kwargs_rmax(self):
        stog = StoG(**{'Rmax': 25.0})
        self.assertEqual(stog.rmax, 25.0)

    def test_stog_init_kwargs_rdelta(self):
        stog = StoG(**{'Rdelta': 0.5})
        self.assertEqual(stog.rdelta, 0.5)

    def test_stog_init_kwargs_rpoints(self):
        stog = StoG(**{'Rpoints': 250})
        self.assertEqual(stog.rdelta, 0.2)

    def test_stog_init_kwargs_density(self):
        stog = StoG(**{'NumberDensity': 2.0})
        self.assertEqual(stog.density, 2.0)

    def test_stog_init_kwargs_low_q_correction(self):
        stog = StoG(**{'OmittedXrangeCorrection': True})
        self.assertEqual(stog.low_q_correction, True)

    def test_stog_init_kwargs_lorch_flag(self):
        stog = StoG(**{'LorchFlag': True})
        self.assertEqual(stog.lorch_flag, True)

    def test_stog_init_kwargs_fourier_filter_cutoff(self):
        stog = StoG(**{'FourierFilter': {'Cutoff': 1.0}})
        self.assertEqual(stog.fourier_filter_cutoff, 1.0)

    def test_stog_init_kwargs_bcoh_sqrd(self):
        stog = StoG(**{'<b_coh>^2': 0.5})
        self.assertEqual(stog.bcoh_sqrd, 0.5)

    def test_stog_init_kwargs_btot_sqrd(self):
        stog = StoG(**{'<b_tot^2>': 0.75})
        self.assertEqual(stog.btot_sqrd, 0.75)

    def test_stog_init_kwargs_qmin_only(self):
        stog = StoG(**{'Merging': {'Transform': {'Qmin': 0.5}}})
        self.assertEqual(stog.qmin, 0.5)
        self.assertEqual(stog.qmax, None)

    def test_stog_init_kwargs_qmax_only(self):
        stog = StoG(**{'Merging': {'Transform': {'Qmax': 30.0}}})
        self.assertEqual(stog.qmin, None)
        self.assertEqual(stog.qmax, 30.0)

    def test_stog_init_kwargs_qmin_and_qmax(self):
        stog = StoG(**{'Merging': {'Transform': {'Qmin': 0.5, 'Qmax': 30.0}}})
        self.assertEqual(stog.qmin, 0.5)
        self.assertEqual(stog.qmax, 30.0)

    def test_stog_init_kwargs_output_stem_name(self):
        stog = StoG(**{'Outputs': {'StemName': 'myName'}})
        self.assertEqual(stog.stem_name, 'myName')


class TestStogAttributes(TestStogBase):
    def setUp(self):
        super(TestStogAttributes, self).setUp()

    def test_stog_xmin_setter(self):
        stog = StoG()
        stog.xmin = 0.25
        self.assertEqual(stog.xmin, 0.25)

    def test_stog_xmax_setter(self):
        stog = StoG()
        stog.xmax = 10.0
        self.assertEqual(stog.xmax, 10.0)

    def test_stog_dr_getter(self):
        stog = StoG()
        self.assertAlmostEqual(stog.dr[0], 0.0)
        self.assertAlmostEqual(stog.dr[-1], 50.0)

    def test_stog_dr_setter_rmax(self):
        stog = StoG()
        stog.rmax = 25.0
        self.assertAlmostEqual(stog.dr[0], 0.0)
        self.assertAlmostEqual(stog.dr[-1], 25.0)

    def test_stog_dr_setter_rmin(self):
        stog = StoG()
        stog.rmin = 10.0
        self.assertAlmostEqual(stog.dr[0], 10.0)
        self.assertAlmostEqual(stog.dr[-1], 50.0)

    def test_stog_dr_setter_rdelta(self):
        stog = StoG()
        stog.rdelta = 0.5
        self.assertEqual(stog.dr[1] - stog.dr[0], 0.5)

    def test_stog_gr_title_function_setter(self):
        stog = StoG()
        stog.gr_title = "G(r) dog"
        self.assertEqual(stog.gr_title, "G(r) dog")

    def test_stog_gr_ft_title_function_setter(self):
        stog = StoG()
        stog.gr_ft_title = "G(r) FT dog"
        self.assertEqual(stog.gr_ft_title, "G(r) FT dog")

    def test_stog_gr_lorch_title_function_setter(self):
        stog = StoG()
        stog.gr_lorch_title = "G(r) FT Lorch dog"
        self.assertEqual(stog.gr_lorch_title, "G(r) FT Lorch dog")

    def test_stog_GKofR_title_function_setter(self):
        stog = StoG()
        stog.GKofR_title = "GK(r) dog"
        self.assertEqual(stog.GKofR_title, "GK(r) dog")

    def test_stog_sq_title_function_setter(self):
        stog = StoG()
        stog.sq_title = "S(Q) dog"
        self.assertEqual(stog.sq_title, "S(Q) dog")

    def test_stog_fq_title_function_setter(self):
        stog = StoG()
        stog.fq_title = "F(Q) dog"
        self.assertEqual(stog.fq_title, "F(Q) dog")

    def test_stog_qsq_minus_one_title_setter(self):
        stog = StoG()
        stog.qsq_minus_one_title = "Q[S(Q)-1] dog"
        self.assertEqual(stog.qsq_minus_one_title, "Q[S(Q)-1] dog")

    def test_stog_sq_ft_title_setter(self):
        stog = StoG()
        stog.sq_ft_title = "S(Q) FT dog"
        self.assertEqual(stog.sq_ft_title, "S(Q) FT dog")

    def test_stog_real_space_function_setter(self):
        stog = StoG()
        stog.real_space_function = "GK(r)"
        self.assertEqual(stog.real_space_function, "GK(r)")
        self.assertEqual(stog.gr_title, "GK(r) Merged")
        self.assertEqual(stog.real_space_function, "GK(r)")
        self.assertEqual(stog.gr_ft_title, "GK(r) FT")
        self.assertEqual(stog.gr_lorch_title, "GK(r) FT Lorched")

    def test_stog_low_q_correction_exception(self):
        stog = StoG()
        with self.assertRaises(TypeError):
            stog.low_q_correction = 1.0

    def test_stog_lorch_flag_exception(self):
        stog = StoG()
        with self.assertRaises(TypeError):
            stog.lorch_flag = 1.0

    def test_stog_real_space_function_exception(self):
        stog = StoG()
        with self.assertRaises(ValueError):
            stog.real_space_function = "Dog"


class TestStogStorageArrays(TestStogBase):
    def setUp(self):
        super(TestStogStorageArrays, self).setUp()
        self.target = np.random.randn(3, 10)

    def test_stog_reciprocal_individuals_setter(self):
        stog = StoG()
        stog.reciprocal_individuals = self.target
        np.testing.assert_allclose(stog.reciprocal_individuals, self.target)

    def test_stog_sq_individuals_setter(self):
        stog = StoG()
        stog.sq_individuals = self.target
        np.testing.assert_allclose(stog.sq_individuals, self.target)

    def test_stog_q_master_setter(self):
        stog = StoG()
        stog.q_master = self.target
        np.testing.assert_allclose(stog.q_master, self.target)

    def test_stog_sq_master_setter(self):
        stog = StoG()
        stog.sq_master = self.target
        np.testing.assert_allclose(stog.sq_master, self.target)

    def test_stog_r_master_setter(self):
        stog = StoG()
        stog.r_master = self.target
        np.testing.assert_allclose(stog.r_master, self.target)

    def test_stog_gr_master_setter(self):
        stog = StoG()
        stog.gr_master = self.target
        np.testing.assert_allclose(stog.gr_master, self.target)


class TestStogGeneralMethods(TestStogBase):
    def setUp(self):
        super(TestStogGeneralMethods, self).setUp()

    def test_stog_append_file(self):
        stog = StoG(**{'Files': ['file1.txt', 'file2.txt']})
        stog.append_file('file3.txt')
        self.assertEqual(stog.files, ['file1.txt', 'file2.txt', 'file3.txt'])

    def test_stog_extend_file_list(self):
        stog = StoG(**{'Files': ['file1.txt', 'file2.txt']})
        stog.extend_file_list(['file3.txt', 'file4.txt'])
        self.assertEqual(
            stog.files, [
                'file1.txt', 'file2.txt', 'file3.txt', 'file4.txt'])


class TestStogDatasetSpecificMethods(TestStogBase):
    def setUp(self):
        super(TestStogDatasetSpecificMethods, self).setUp()

    def test_stog_apply_scales_and_offset(self):
        q, sq, dq = StoG.apply_scales_and_offset(self.q, self.sq)
        np.testing.assert_allclose(q, self.q)
        np.testing.assert_allclose(sq, self.sq)
        np.testing.assert_allclose(dq, np.zeros_like(sq))

    def test_stog_apply_scales_and_offset_with_dy(self):
        q, sq, dq = StoG.apply_scales_and_offset(self.q, self.sq, dy=self.sq)
        np.testing.assert_allclose(q, self.q)
        np.testing.assert_allclose(sq, self.sq)
        np.testing.assert_allclose(dq, self.sq)

    def test_stog_apply_scales_and_offset_with_yscale(self):
        q, sq, dq = StoG.apply_scales_and_offset(
            self.q, self.sq, dy=self.sq, yscale=2.0)
        np.testing.assert_allclose(q, self.q)
        np.testing.assert_allclose(sq, self.sq * 2.0)
        np.testing.assert_allclose(dq, self.sq * 2.0)

    def test_stog_apply_scales_and_offset_with_yoffset(self):
        q, sq, dq = StoG.apply_scales_and_offset(
            self.q, self.sq, dy=self.sq, yoffset=2.0)
        np.testing.assert_allclose(q, self.q)
        np.testing.assert_allclose(sq, self.sq + 2.0)
        np.testing.assert_allclose(dq, self.sq)

    def test_stog_apply_scales_and_offset_with_xoffset(self):
        q, sq, dq = StoG.apply_scales_and_offset(
            self.q, self.sq, dy=self.sq, xoffset=2.0)
        np.testing.assert_allclose(q, self.q + 2.0)
        np.testing.assert_allclose(sq, self.sq)
        np.testing.assert_allclose(dq, self.sq)

    def test_stog_add_dataset(self):
        # Number of decimal places for precision
        places = 5

        # Initialize with material info for Argon
        stog = StoG(**{'<b_coh>^2': self.kwargs['<b_coh>^2'],
                       '<b_tot^2>': self.kwargs['<b_tot^2>']})

        # Add the S(Q) data set and check values against targets
        info = {
            'data': [self.q, self.sq],
            'ReciprocalFunction': 'S(Q)'}
        stog.add_dataset(info)
        self.assertEqual(
            stog.reciprocal_individuals[0][self.first],
            self.reciprocal_xtarget)
        self.assertAlmostEqual(
            stog.reciprocal_individuals[1][self.first],
            self.sq_target[0],
            places=places)
        self.assertEqual(stog.reciprocal_individuals[2][self.first], 0.0)
        self.assertAlmostEqual(
            stog.sq_individuals[1][self.first],
            self.sq_target[0],
            places=places)
        self.assertEqual(stog.sq_individuals[2][self.first], 0.0)

        # Add the Q[S(Q)-1] data set and check values for it and S(Q) against
        # targets
        stride = stog.reciprocal_individuals.shape[1]
        info = {
            'data': [self.q, self.fq],
            'ReciprocalFunction': 'Q[S(Q)-1]'
        }
        stog.add_dataset(info)

        self.assertAlmostEqual(
            stog.reciprocal_individuals[1][self.first + stride],
            self.fq_target[0],
            places=places)
        self.assertEqual(
            stog.reciprocal_individuals[2][self.first + stride],
            0.0)
        self.assertAlmostEqual(
            stog.sq_individuals[1][self.first],
            self.sq_target[0],
            places=places)
        self.assertEqual(stog.sq_individuals[2][self.first], 0.0)

        # Add the FK(Q) data set and check values for it and S(Q) against
        # targets
        stride = stog.reciprocal_individuals.shape[1]
        info = {
            'data': [self.q, self.fq_keen],
            'ReciprocalFunction': 'FK(Q)'}
        stog.add_dataset(info)
        self.assertAlmostEqual(
            stog.reciprocal_individuals[1][self.first + stride],
            self.fq_keen_target[0],
            places=places)
        self.assertEqual(
            stog.reciprocal_individuals[2][self.first + stride],
            0.0)
        self.assertAlmostEqual(
            stog.sq_individuals[1][self.first],
            self.sq_target[0],
            places=places)
        self.assertEqual(stog.sq_individuals[2][self.first], 0.0)

        # Add the DCS(Q) data set and check values for it and S(Q) against
        # targets
        stride = stog.reciprocal_individuals.shape[1]
        info = {
            'data': [self.q, self.dcs],
            'ReciprocalFunction': 'DCS(Q)'}
        stog.add_dataset(info)
        self.assertAlmostEqual(
            stog.reciprocal_individuals[1][self.first + stride],
            self.dcs_target[0],
            places=places)
        self.assertEqual(
            stog.reciprocal_individuals[2][self.first + stride],
            0.0)
        self.assertAlmostEqual(
            stog.sq_individuals[1][self.first],
            self.sq_target[0],
            places=places)
        self.assertEqual(stog.sq_individuals[2][self.first], 0.0)

    def test_stog_add_dataset_yscale(self):
        # Scale S(Q) and make sure it does not equal original target values
        stog = StoG()
        info = {
            'data': [self.q, self.sq],
            'ReciprocalFunction': 'S(Q)',
            'Y': {'Scale': 2.0}}
        stog.add_dataset(info)
        self.assertNotEqual(
            stog.reciprocal_individuals[1][self.first],
            self.sq_target[0])
        self.assertEqual(stog.reciprocal_individuals[2][self.first], 0.0)
        self.assertNotEqual(
            stog.sq_individuals[1][self.first],
            self.sq_target[0])
        self.assertEqual(stog.sq_individuals[2][self.first], 0.0)

    def test_stog_add_dataset_yoffset(self):
        # Offset S(Q) and make sure it does not equal original target values
        stog = StoG()
        info = {
            'data': [self.q, self.sq],
            'ReciprocalFunction': 'S(Q)',
            'Y': {'Offset': 2.0}}
        stog.add_dataset(info)
        self.assertNotEqual(
            stog.reciprocal_individuals[1][self.first],
            self.sq_target[0])
        self.assertEqual(stog.reciprocal_individuals[2][self.first], 0.0)
        self.assertNotEqual(
            stog.sq_individuals[1][self.first],
            self.sq_target[0])
        self.assertEqual(stog.sq_individuals[2][self.first], 0.0)

    def test_stog_add_dataset_yscale_with_dy(self):
        # Scale S(Q) and make sure it does not equal original target values
        stog = StoG()
        info = {
            'data': [self.q, self.sq, self.sq],
            'ReciprocalFunction': 'S(Q)',
            'Y': {'Scale': 2.0}}
        stog.add_dataset(info)

        self.assertNotEqual(
            stog.reciprocal_individuals[1][self.first],
            self.sq_target[0])

        dsq_target = info['Y']['Scale'] * 2.59173
        self.assertEqual(
            stog.reciprocal_individuals[2][self.first],
            dsq_target)

        self.assertNotEqual(
            stog.sq_individuals[1][self.first],
            self.sq_target[0])

        self.assertEqual(stog.sq_individuals[2][self.first], dsq_target)

    def test_stog_add_dataset_yoffset_with_dy(self):
        # Offset S(Q) and make sure it does not equal original target values
        stog = StoG()
        info = {
            'data': [self.q, self.sq, self.sq],
            'ReciprocalFunction': 'S(Q)',
            'Y': {'Offset': 2.0}}
        stog.add_dataset(info)

        self.assertNotEqual(
            stog.reciprocal_individuals[1][self.first],
            self.sq_target[0])

        dsq_target = 2.59173
        self.assertEqual(
            stog.reciprocal_individuals[2][self.first],
            dsq_target)

        self.assertNotEqual(
            stog.sq_individuals[1][self.first],
            self.sq_target[0])

        self.assertEqual(stog.sq_individuals[2][self.first], dsq_target)

    def test_stog_add_dataset_xoffset(self):
        # Offset Q from 1.96 -> 2.14
        stog = StoG()
        info = {
            'data': [self.q, self.sq],
            'ReciprocalFunction': 'S(Q)',
            'X': {'Offset': 0.2}}
        stog.add_dataset(info)
        self.assertEqual(stog.reciprocal_individuals[0][self.first], 2.14)

    def test_stog_add_dataset_qmin_qmax_crop(self):
        # Check qmin and qmax apply cropping
        stog = StoG()
        info = {'data': [self.q, self.sq],
                'ReciprocalFunction': 'S(Q)'}
        stog.qmin = 1.5
        stog.qmax = 12.0
        stog.add_dataset(info)
        self.assertEqual(stog.reciprocal_individuals[0][0], stog.qmin)
        self.assertEqual(stog.reciprocal_individuals[0][-1], stog.qmax)

    def test_stog_add_dataset_default_reciprocal_space_function(self):
        # Checks the default reciprocal space function is S(Q)
        places = 5
        stog = StoG()
        info = {'data': [self.q, self.sq]}
        stog.add_dataset(info)
        self.assertEqual(
            stog.reciprocal_individuals[0][self.first],
            self.reciprocal_xtarget)
        self.assertEqual(stog.reciprocal_individuals[2][self.first], 0.0)
        self.assertAlmostEqual(
            stog.reciprocal_individuals[1][self.first],
            self.sq_target[0],
            places=places)
        self.assertEqual(stog.sq_individuals[2][self.first], 0.0)

    def test_stog_add_dataset_wrong_reciprocal_space_function_exception(self):
        # Check qmin and qmax apply cropping
        stog = StoG()
        info = {'data': [self.q, self.sq],
                'ReciprocalFunction': 'ABCDEFG(Q)'}
        with self.assertRaises(ValueError):
            stog.add_dataset(info)

    def test_stog_read_dataset(self):
        # Number of decimal places for precision
        places = 5

        # Load S(Q) for Argon from test data
        stog = StoG(**{'<b_coh>^2': self.kwargs['<b_coh>^2'],
                       '<b_tot^2>': self.kwargs['<b_tot^2>']})
        info = {
            'Filename': get_data_path(
                self.material.reciprocal_space_filename),
            'ReciprocalFunction': 'S(Q)',
            'Qmin': 0.02,
            'Qmax': 35.2,
            'Y': {
                'Offset': 0.0,
                'Scale': 1.0},
            'X': {
                'Offset': 0.0}}

        info['index'] = 0
        stog.read_dataset(info)

        # Check S(Q) data against targets
        self.assertEqual(
            stog.reciprocal_individuals[0][self.first],
            self.reciprocal_xtarget)
        self.assertAlmostEqual(
            stog.reciprocal_individuals[1][self.first],
            self.sq_target[0],
            places=places)
        self.assertEqual(
            stog.reciprocal_individuals[2][self.first],
            self.dsq_target)
        self.assertAlmostEqual(
            stog.sq_individuals[1][self.first],
            self.sq_target[0],
            places=places)
        self.assertEqual(
            stog.sq_individuals[2][self.first],
            self.dsq_target)

    def test_stog_read_dataset_xcol_data_format_exception(self):
        stog = StoG()
        filename = get_data_path(self.material.reciprocal_space_filename)
        with self.assertRaises(RuntimeError):
            stog.read_dataset({'Filename': filename}, xcol=99)

    def test_stog_read_dataset_ycol_data_format_exception(self):
        stog = StoG()
        filename = get_data_path(self.material.reciprocal_space_filename)
        with self.assertRaises(RuntimeError):
            stog.read_dataset({'Filename': filename}, ycol=99)

    def test_stog_read_dataset_dycol_too_large(self):
        # Number of decimal places for precision
        places = 5

        # Load S(Q) for Argon from test data
        stog = StoG(**{'<b_coh>^2': self.kwargs['<b_coh>^2'],
                       '<b_tot^2>': self.kwargs['<b_tot^2>']})
        info = {
            'Filename': get_data_path(
                self.material.reciprocal_space_filename),
            'ReciprocalFunction': 'S(Q)',
            'Qmin': 0.02,
            'Qmax': 35.2,
            'Y': {
                'Offset': 0.0,
                'Scale': 1.0},
            'X': {
                'Offset': 0.0}}

        info['index'] = 0
        stog.read_dataset(info, dycol=99)

        # Check S(Q) data against targets
        self.assertEqual(
            stog.reciprocal_individuals[0][self.first],
            self.reciprocal_xtarget)
        self.assertAlmostEqual(
            stog.reciprocal_individuals[1][self.first],
            self.sq_target[0],
            places=places)
        self.assertEqual(
            stog.reciprocal_individuals[2][self.first],
            0.0)
        self.assertAlmostEqual(
            stog.sq_individuals[1][self.first],
            self.sq_target[0],
            places=places)

    def test_stog_read_all_data_assertion(self):
        stog = StoG()
        with self.assertRaises(NoInputFilesException):
            stog.read_all_data()

        stog.files = list()
        with self.assertRaises(NoInputFilesException):
            stog.read_all_data()

    def test_stog_read_all_data_for_files_length(self):
        # Load S(Q) for Argon from test data
        stog = StoG()
        stog.files = self.kwargs_for_files['Files']
        stog.read_all_data()

        # Check S(Q) data against targets
        self.assertEqual(
            len(stog.files),
            len(self.kwargs_for_files['Files']))

    def test_stog_read_all_data(self):
        # Number of decimal places for precision
        places = 5

        # Load S(Q) for Argon from test data
        stog = StoG()
        stog.files = self.kwargs_for_files['Files']
        stog.read_all_data()

        # Check S(Q) data against targets
        self.assertEqual(
            stog.reciprocal_individuals[0][self.first], self.reciprocal_xtarget)
        for index in range(len(stog.files)):
            self.assertAlmostEqual(
                stog.reciprocal_individuals[1][self.first],
                self.sq_target[0],
                places=places)
            self.assertAlmostEqual(
                stog.sq_individuals[1][self.first],
                self.sq_target[0],
                places=places)

    def test_stog_merge_data(self):
        # Number of decimal places for precision
        places = 5

        # Load S(Q) for Argon from test data
        stog = StoG()
        stog.files = self.kwargs_for_files['Files']
        stog.read_all_data()
        stog.merge_data()

        # Check S(Q) data against targets
        self.assertEqual(
            stog.q_master[stog.sq_title][self.first],
            self.reciprocal_xtarget)
        self.assertAlmostEqual(
            stog.sq_master[stog.sq_title][self.first],
            self.sq_target[0],
            places=places)

    def test_stog_merge_data_qsq_opts_scale(self):
        # Number of decimal places for precision
        places = 5

        # Load S(Q) for Argon from test data
        stog = StoG()
        stog.files = self.kwargs_for_files['Files']
        stog.read_all_data()
        qsq_opts = {'Y': {'Scale': 2.0}}

        # Test Q[S(Q)-1] scale
        stog.merged_opts['Q[S(Q)-1]'] = qsq_opts
        stog.merge_data()
        self.assertAlmostEqual(
            stog.sq_master[stog.qsq_minus_one_title][self.first],
            qsq_opts['Y']['Scale'] * self.fq_target[0],
            places=places)

    def test_stog_merge_data_qsq_opts_offset(self):
        # Number of decimal places for precision
        places = 5

        # Load S(Q) for Argon from test data
        stog = StoG()
        stog.files = self.kwargs_for_files['Files']
        stog.read_all_data()
        qsq_opts = {'Y': {'Offset': 1.0}}

        # Test Q[S(Q)-1] scale
        stog.merged_opts['Q[S(Q)-1]'] = qsq_opts
        stog.merge_data()
        self.assertAlmostEqual(
            stog.sq_master[stog.qsq_minus_one_title][self.first],
            qsq_opts['Y']['Offset'] + self.fq_target[0],
            places=places)


class TestStogTransformSpecificMethods(TestStogDatasetSpecificMethods):
    def setUp(self):
        super(TestStogTransformSpecificMethods, self).setUp()
        self.lowR_target = 0.4720653

    def test_stog_transform_merged_default(self):
        # Number of decimal places for precision
        places = 2

        # Load S(Q) for Argon from test data
        stog = StoG(**self.kwargs_for_stog_input)
        stog.files = self.kwargs_for_files['Files']
        stog.read_all_data()
        stog.merge_data()
        stog.transform_merged()

        # Check g(r) data against targets
        self.assertAlmostEqual(
            stog.r_master[stog.gr_title][self.real_space_first],
            self.real_xtarget,
            places=places)
        self.assertAlmostEqual(
            stog.gr_master[stog.gr_title][self.real_space_first],
            self.gofr_target[0],
            places=places)

        # Test if no dr defined
        stog.dr = None
        stog.transform_merged()

        # Check g(r) data against targets
        self.assertAlmostEqual(
            stog.r_master[stog.gr_title][self.real_space_first],
            self.real_xtarget,
            places=places)
        self.assertAlmostEqual(
            stog.gr_master[stog.gr_title][self.real_space_first],
            self.gofr_target[0],
            places=places)

    def test_stog_transform_merged_GofR(self):
        # Number of decimal places for precision
        places = 2

        # Load S(Q) for Argon from test data
        stog = StoG(**self.kwargs_for_stog_input)
        stog.files = self.kwargs_for_files['Files']
        stog.real_space_function = "G(r)"
        stog.read_all_data()
        stog.merge_data()
        stog.transform_merged()

        # Check G(r) data against targets
        self.assertAlmostEqual(
            stog.r_master[stog.gr_title][self.real_space_first],
            self.real_xtarget,
            places=places)
        self.assertAlmostEqual(
            stog.gr_master[stog.gr_title][self.real_space_first],
            self.GofR_target[0],
            places=places)

    def test_stog_transform_merged_GKofR(self):
        # Number of decimal places for precision
        places = 2

        # Load S(Q) for Argon from test data
        stog = StoG(**self.kwargs_for_stog_input)
        stog.files = self.kwargs_for_files['Files']
        stog.real_space_function = "GK(r)"
        stog.read_all_data()
        stog.merge_data()
        stog.transform_merged()

        # Check GK(r) data against targets
        self.assertAlmostEqual(
            stog.r_master[stog.gr_title][self.real_space_first],
            self.real_xtarget,
            places=places)
        self.assertAlmostEqual(
            stog.gr_master[stog.gr_title][self.real_space_first],
            self.GKofR_target[0],
            places=places)

    def test_stog_transform_merged_for_nan_after_filter(self):
        # Load S(Q) for Argon from test data
        stog = StoG(**self.kwargs_for_stog_input)
        stog.files = self.kwargs_for_files['Files']
        stog.real_space_function = "GK(r)"
        stog.read_all_data()
        stog.merge_data()
        stog.transform_merged()

        self.assertFalse(
            np.isnan(stog.sq_master[stog.sq_title]).any())
        self.assertFalse(
            np.isnan(stog.gr_master[stog.gr_title]).any())

    def test_stog_fourier_filter(self):
        # Number of decimal places for precision
        places = 1

        # Load S(Q) for Argon from test data
        stog = StoG(**self.kwargs_for_stog_input)
        stog.files = self.kwargs_for_files['Files']
        stog.read_all_data()
        stog.merge_data()
        stog.transform_merged()
        stog.fourier_filter()

        # Check g(r) data against targets
        self.assertAlmostEqual(
            stog.r_master[stog.gr_ft_title][self.real_space_first],
            self.real_xtarget,
            places=places)
        self.assertAlmostEqual(
            stog.gr_master[stog.gr_ft_title][self.real_space_first],
            self.gofr_ff_target[0],
            places=places)

    def test_stog_fourier_filter_before_transform_merged_call(self):
        # Number of decimal places for precision
        places = 1

        # Load S(Q) for Argon from test data
        stog = StoG(**self.kwargs_for_stog_input)
        stog.files = self.kwargs_for_files['Files']
        stog.read_all_data()
        stog.merge_data()
        stog.fourier_filter()

        # Check g(r) data against targets
        self.assertAlmostEqual(
            stog.r_master[stog.gr_ft_title][self.real_space_first],
            self.real_xtarget,
            places=places)
        self.assertAlmostEqual(
            stog.gr_master[stog.gr_ft_title][self.real_space_first],
            self.gofr_ff_target[0],
            places=places)

    def test_stog_fourier_filter_GofR(self):
        # Number of decimal places for precision
        places = 1

        # Load S(Q) for Argon from test data
        stog = StoG(**self.kwargs_for_stog_input)
        stog.files = self.kwargs_for_files['Files']
        stog.real_space_function = "G(r)"
        stog.read_all_data()
        stog.merge_data()
        stog.transform_merged()
        stog.fourier_filter()

        # Check g(r) data against targets
        self.assertAlmostEqual(
            stog.r_master[stog.gr_ft_title][self.real_space_first],
            self.real_xtarget,
            places=places)
        self.assertAlmostEqual(
            stog.gr_master[stog.gr_ft_title][self.real_space_first],
            self.GofR_ff_target[0],
            places=places)

    def test_stog_fourier_filter_GKofR(self):
        # Number of decimal places for precision
        places = 1

        # Load S(Q) for Argon from test data
        stog = StoG(**self.kwargs_for_stog_input)
        stog.files = self.kwargs_for_files['Files']
        stog.real_space_function = "GK(r)"
        stog.read_all_data()
        stog.merge_data()
        stog.transform_merged()
        stog.fourier_filter()

        # Check g(r) data against targets
        self.assertAlmostEqual(
            stog.r_master[stog.gr_ft_title][self.real_space_first],
            self.real_xtarget,
            places=places)
        self.assertAlmostEqual(
            stog.gr_master[stog.gr_ft_title][self.real_space_first],
            self.GKofR_ff_target[0],
            places=places)

    def test_stog_fourier_filter_for_nan_after_filter(self):
        # Load S(Q) for Argon from test data
        stog = StoG(**self.kwargs_for_stog_input)
        stog.files = self.kwargs_for_files['Files']
        stog.read_all_data()
        stog.merge_data()
        stog.transform_merged()
        stog.fourier_filter()

        self.assertFalse(
            np.isnan(stog.gr_master[stog.gr_title]).any())
        self.assertFalse(
            np.isnan(stog.sq_master[stog.sq_title]).any())
        self.assertFalse(
            np.isnan(stog.sq_master[stog._ft_title]).any())
        self.assertFalse(
            np.isnan(stog.sq_master[stog.sq_ft_title]).any())

    def test_stog_apply_lorch_default(self):
        # Number of decimal places for precision
        places = 5

        # Load S(Q) for Argon from test data
        stog = StoG(**self.kwargs_for_stog_input)
        stog.files = self.kwargs_for_files['Files']
        stog.read_all_data()
        stog.merge_data()
        stog.transform_merged()
        q, sq, r, gr = stog.fourier_filter()
        stog.apply_lorch(q, sq, r)

        self.assertAlmostEqual(
            stog.gr_master[stog.gr_lorch_title][self.real_space_first],
            self.gofr_lorch_target[0],
            places=places)

    def test_stog_apply_lorch_GofR(self):
        # Number of decimal places for precision
        places = 5

        # Load S(Q) for Argon from test data
        stog = StoG(**self.kwargs_for_stog_input)
        stog.files = self.kwargs_for_files['Files']
        stog.real_space_function = "G(r)"
        stog.read_all_data()
        stog.merge_data()
        stog.transform_merged()
        q, sq, r, gr = stog.fourier_filter()
        stog.apply_lorch(q, sq, r)

        self.assertAlmostEqual(
            stog.gr_master[stog.gr_lorch_title][self.real_space_first],
            self.GofR_lorch_target[0],
            places=places)

    def test_stog_apply_lorch_GKofR(self):
        # Number of decimal places for precision
        places = 5

        # Load S(Q) for Argon from test data
        stog = StoG(**self.kwargs_for_stog_input)
        stog.files = self.kwargs_for_files['Files']
        stog.real_space_function = "GK(r)"
        stog.read_all_data()
        stog.merge_data()
        stog.transform_merged()
        q, sq, r, gr = stog.fourier_filter()
        stog.apply_lorch(q, sq, r)

        self.assertAlmostEqual(
            stog.gr_master[stog.gr_lorch_title][self.real_space_first],
            self.GKofR_lorch_target[0], places=places)

    def test_stog_lowR_mean_square(self):
        # Load S(Q) for Argon from test data
        stog = StoG(**self.kwargs_for_stog_input)
        stog.files = self.kwargs_for_files['Files']
        stog.read_all_data()
        stog.merge_data()
        stog.transform_merged()
        gr = stog.gr_master[stog.gr_title]
        cost = stog._lowR_mean_square(stog.dr, gr)
        self.assertAlmostEqual(cost, self.lowR_target, places=7)

    def test_stog_get_lowR_mean_square(self):
        # Load S(Q) for Argon from test data
        stog = StoG(**self.kwargs_for_stog_input)
        stog.files = self.kwargs_for_files['Files']
        stog.read_all_data()
        stog.merge_data()
        stog.transform_merged()
        q, sq, r, gr = stog.fourier_filter()
        cost = stog._get_lowR_mean_square()
        self.assertAlmostEqual(cost, self.lowR_target, places=7)


class TestStogBookkeepingMethods(TestStogDatasetSpecificMethods):
    def setUp(self):
        super(TestStogBookkeepingMethods, self).setUp()
        stog = StoG(**self.kwargs_for_stog_input)
        stog.files = self.kwargs_for_files['Files']
        stog.read_all_data()
        stog.merge_data()
        stog.transform_merged()

        self.stog = stog

    def test_stog_add_keen_fq(self):
        stog = self.stog
        q = stog.sq_master[stog.sq_title]
        sq = stog.sq_master[stog.sq_title]
        stog._add_keen_fq(q, sq)
        self.assertTrue(stog.fq_title in stog.sq_master)

    def test_stog_add_keen_gr_default(self):
        stog = self.stog
        r = stog.gr_master[stog.gr_title]
        gr = stog.gr_master[stog.gr_title]
        stog._add_keen_gr(r, gr)
        self.assertTrue(stog.GKofR_title in stog.gr_master)

    def test_stog_add_keen_gr_GofR(self):
        stog = StoG(**self.kwargs_for_stog_input)
        stog.real_space_function = "G(r)"
        stog.files = self.kwargs_for_files['Files']
        stog.read_all_data()
        stog.merge_data()
        stog.transform_merged()
        r = stog.gr_master[stog.gr_title]
        gr = stog.gr_master[stog.gr_title]
        stog._add_keen_gr(r, gr)
        self.assertTrue(stog.GKofR_title in stog.gr_master)

    def test_stog_add_keen_gr_GKofR(self):
        stog = StoG(**self.kwargs_for_stog_input)
        stog.real_space_function = "GK(r)"
        stog.files = self.kwargs_for_files['Files']
        stog.read_all_data()
        stog.merge_data()
        stog.transform_merged()
        r = stog.gr_master[stog.gr_title]
        gr = stog.gr_master[stog.gr_title]
        stog._add_keen_gr(r, gr)
        self.assertTrue(stog.GKofR_title in stog.gr_master)


class TestStogOutputMethods(TestStogDatasetSpecificMethods):
    def setUp(self):
        super(TestStogOutputMethods, self).setUp()
        stog = StoG(**self.kwargs_for_stog_input)
        stog.stem_name = "dog"
        stog.files = self.kwargs_for_files['Files']
        stog.read_all_data()
        stog.merge_data()
        stog.transform_merged()

        self.stog = stog

    # Decorator to provide the data to run each write_out_<type> test
    def data_provider(self, x, y, filename):
        def write_out_decorator(write_out_func):
            def wrap_function(*args):
                # Using stem name
                write_out_func()

                data = {}
                data['x'], data['y'] = np.genfromtxt(
                    filename,
                    skip_header=2,
                    unpack=True)

                np.testing.assert_allclose(data['x'], x)
                np.testing.assert_allclose(
                    data['y'] - y,
                    np.zeros(
                        len(y)),
                    rtol=2.0,
                    atol=2.0,
                    equal_nan=True)

                os.remove(filename)

                # Using set filename
                tmp_filename = tempfile.mkstemp()[1]

                write_out_func(filename=tmp_filename)

                data['x'], data['y'] = np.genfromtxt(
                    tmp_filename,
                    skip_header=2,
                    unpack=True)

                np.testing.assert_allclose(data['x'], x)
                np.testing.assert_allclose(data['y'], y)
                os.remove(tmp_filename)

            return wrap_function
        return write_out_decorator

    # Tests
    def test_stog_write_df(self):
        outfile_path = tempfile.mkstemp()[1]
        x = self.stog.q_master[self.stog.sq_title]
        y = self.stog.sq_master[self.stog.sq_title]
        self.stog._write_out_to_file(x, y, outfile_path)
        data = {}
        data['x'], data['y'] = np.genfromtxt(
            outfile_path,
            skip_header=2,
            unpack=True)

        q = self.stog.q_master[self.stog.sq_title]
        sq = self.stog.sq_master[self.stog.sq_title]

        np.testing.assert_allclose(data['x'], q)
        np.testing.assert_allclose(data['y'], sq)
        os.remove(outfile_path)

    def test_write_out_merged_sq(self):
        # Have to decorate after the setUp() is called for the self.* args to
        # work
        @self.data_provider(
            self.stog.q_master[self.stog.sq_title],
            self.stog.sq_master[self.stog.sq_title],
            "dog.sq")
        def decorated_write_out_merged(*args, **kwargs):
            self.stog.write_out_merged_sq(*args, **kwargs)
        decorated_write_out_merged()

    def test_write_out_merged_gr(self):
        # Have to decorate after the setUp() is called for the self.* args to
        # work
        @self.data_provider(
            self.stog.r_master[self.stog.gr_title],
            self.stog.gr_master[self.stog.gr_title],
            "dog.gr")
        def decorated_write_out_merged(*args, **kwargs):
            self.stog.write_out_merged_gr(*args, **kwargs)
        decorated_write_out_merged()

    def test_write_out_ft_sq(self):
        self.stog.fourier_filter()
        # Have to decorate after the setUp() is called for the self.* args to
        # work

        @self.data_provider(
            self.stog.q_master[self.stog.sq_ft_title],
            self.stog.sq_master[self.stog.sq_ft_title],
            'dog_ft.sq')
        def decorated_write_out_merged(*args, **kwargs):
            self.stog.write_out_ft_sq(*args, **kwargs)
        decorated_write_out_merged()


if __name__ == '__main__':
    unittest.main()  # pragma: no cover



"""
==========
Converter
==========

This module defines the Converter class
that converts functions in the same space
"""

import numpy as np


class Converter:
    """
    The Converter class is used to convert between
    either different reciprocal space functions or
    different real space functions

    :examples:

    >>> import numpy
    >>> from pystog import Converter
    >>> converter = Converter()
    >>> q, sq = numpy.loadtxt("my_sofq_file.txt",unpack=True)
    >>> fq, dfq = converter.S_to_F(q, sq)
    >>> r, gr = numpy.loadtxt("my_gofr_file.txt",unpack=True)
    >>> kwargs = {'rho' : 1.0}
    >>> gr_keen, dgr_keen = converter.g_to_GK(r, gr, **kwargs)
    """

    def __init__(self):
        pass

    def _safe_divide(self, numerator, denominator):
        numerator = np.array(numerator)
        denominator = np.array(denominator)
        mask = (denominator > 0.0)
        out = np.zeros_like(numerator)
        out[mask] = numerator[mask] / denominator[mask]
        return out

    # Reciprocal Space Conversions

    def F_to_S(self, q, fq, dfq=None, **kwargs):
        """
        Converts from :math:`Q[S(Q)-1]` to :math:`S(Q)`

        :param q: :math:`Q`-space vector
        :type q: numpy.array or list
        :param fq: :math:`Q[S(Q)-1]` vector
        :type fq: numpy.array or list
        :param dfq: uncertainty vector
        :type dfq: numpy.array or list

        :return: (:math:`S(Q)` vector, uncertainty vector)
        :rtype: (numpy.array, numpy.array)
        """
        if dfq is None:
            dfq = np.zeros_like(fq)
        return (self._safe_divide(fq, q) + 1., self._safe_divide(dfq, q))

    def F_to_FK(self, q, fq, dfq=None, **kwargs):
        """
        Converts from :math:`Q[S(Q)-1]` to :math:`F(Q)`

        :param q: :math:`Q`-space vector
        :type q: numpy.array or list
        :param fq: :math:`Q[S(Q)-1]` vector
        :type fq: numpy.array or list
        :param dfq: uncertainty vector
        :type dfq: numpy.array or list

        :return: (:math:`F(Q)` vector, uncertainty vector)
        :rtype: (numpy.array, numpy.array)
        """
        if dfq is None:
            dfq = np.zeros_like(fq)
        return (kwargs['<b_coh>^2'] * self._safe_divide(fq, q),
                kwargs['<b_coh>^2'] * self._safe_divide(dfq, q))

    def F_to_DCS(self, q, fq, dfq=None, **kwargs):
        """
        Converts from :math:`Q[S(Q)-1]` to
        :math:`\\frac{d \\sigma}{d \\Omega}(Q)`

        :param q: :math:`Q`-space vector
        :type q: numpy.array or list
        :param fq: :math:`Q[S(Q)-1]` vector
        :type fq: numpy.array or list
        :param dfq: uncertainty vector
        :type dfq: numpy.array or list

        :return: (:math:`\\frac{d \\sigma}{d \\Omega}(Q)` vector,
                 uncertainty vector)
        :rtype: (numpy.array, numpy.array)
        """
        fq, dfq = self.F_to_FK(q, fq, dfq, **kwargs)
        return self.FK_to_DCS(q, fq, dfq, **kwargs)

    # S(Q)
    def S_to_F(self, q, sq, dsq=None, **kwargs):
        """
        Convert :math:`S(Q)` to :math:`Q[S(Q)-1]`

        :param q: :math:`Q`-space vector
        :type q: numpy.array or list
        :param sq: :math:`S(Q)` vector
        :type sq: numpy.array or list
        :param dfq: uncertainty vector
        :type dfq: numpy.array or list
        :param dsq: uncertainty vector
        :type dsq: numpy.array or list

        :return: (:math:`Q[S(Q)-1]` vector, uncertainty vector)
        :rtype: (numpy.array, numpy.array)
        """
        if dsq is None:
            dsq = np.zeros_like(sq)
        return (q * (sq - 1.), q * dsq)

    def S_to_FK(self, q, sq, dsq=None, **kwargs):
        """
        Convert :math:`S(Q)` to :math:`F(Q)`

        :param q: :math:`Q`-space vector
        :type q: numpy.array or list
        :param sq: :math:`S(Q)` vector
        :type sq: numpy.array or list
        :param dsq: uncertainty vector
        :type dsq: numpy.array or list

        :return: (:math:`F(Q)` vector, uncertainty vector)
        :rtype: (numpy.array, numpy.array)
        """
        fq, dfq = self.S_to_F(q, sq, dsq)
        return self.F_to_FK(q, fq, dfq, **kwargs)

    def S_to_DCS(self, q, sq, dsq=None, **kwargs):
        """
        Convert :math:`S(Q)` to :math:`\\frac{d \\sigma}{d \\Omega}(Q)`

        :param q: :math:`Q`-space vector
        :type q: numpy.array or list
        :param sq: :math:`S(Q)` vector
        :type sq: numpy.array or list
        :param dsq: uncertainty vector
        :type dsq: numpy.array or list

        :return: (:math:`\\frac{d \\sigma}{d \\Omega}(Q)` vector,
                 uncertainty vector)
        :rtype: (numpy.array, numpy.array)
        """
        fq, dfq = self.S_to_FK(q, sq, dsq, **kwargs)
        return self.FK_to_DCS(q, fq, dfq, **kwargs)

    # Keen's F(Q)
    def FK_to_F(self, q, fq_keen, dfq_keen=None, **kwargs):
        """
        Convert :math:`F(Q)` to :math:`Q[S(Q)-1]`

        :param q: :math:`Q`-space vector
        :type q: numpy.array or list
        :param fq_keen: :math:`F(Q)` vector
        :type fq_keen: numpy.array or list
        :param dfq_keen: uncertainty vector
        :type dfq_keen: numpy.array or list

        :return: (:math:`Q[S(Q)-1]` vector, uncertainty vector)
        :rtype: (numpy.array, numpy.array)
        """
        if dfq_keen is None:
            dfq_keen = np.zeros_like(fq_keen)
        return (q * fq_keen / kwargs['<b_coh>^2'],
                q * dfq_keen / kwargs['<b_coh>^2'])

    def FK_to_S(self, q, fq_keen, dfq_keen=None, **kwargs):
        """
        Convert :math:`F(Q)` to :math:`S(Q)`

        :param q: :math:`Q`-space vector
        :type q: numpy.array or list
        :param fq_keen: :math:`F(Q)` vector
        :type fq_keen: numpy.array or list
        :param dfq_keen: uncertainty vector
        :type dfq_keen: numpy.array or list

        :return: (:math:`S(Q)` vector, uncertainty vector)
        :rtype: (numpy.array, numpy.array)
        """
        fq, dfq = self.FK_to_F(q, fq_keen, dfq_keen, **kwargs)
        return self.F_to_S(q, fq, dfq)

    def FK_to_DCS(self, q, fq, dfq=None, **kwargs):
        """
        Convert :math:`F(Q)` to :math:`\\frac{d \\sigma}{d \\Omega}(Q)`

        :param q: :math:`Q`-space vector
        :type q: numpy.array or list
        :param fq: :math:`F(Q)` vector
        :type fq: numpy.array or list
        :param dfq: uncertainty vector
        :type dfq: numpy.array or list

        :return: (:math:`\\frac{d \\sigma}{d \\Omega}(Q)` vector,
                 uncertainty vector)
        :rtype: (numpy.array, numpy.array)
        """
        return fq + kwargs['<b_tot^2>'], dfq

    # Differential cross-section = d_simga / d_Omega
    def DCS_to_F(self, q, dcs, ddcs=None, **kwargs):
        """
        Convert :math:`\\frac{d \\sigma}{d \\Omega}(Q)` to :math:`Q[S(Q)-1]`

        :param q: Q-space vector
        :type q: numpy.array or list
        :param dcs: :math:`\\frac{d \\sigma}{d \\Omega}(Q)` vector
        :type dcs: numpy.array or list
        :param ddcs: uncertainty vector
        :type ddcs: numpy.array or list

        :return: (:math:`Q[S(Q)-1]` vector, uncertainty vector)
        :rtype: (numpy.array, numpy.array)
        """
        fq, dfq = self.DCS_to_FK(q, dcs, ddcs, **kwargs)
        return self.FK_to_F(q, fq, dfq, **kwargs)

    def DCS_to_S(self, q, dcs, ddcs=None, **kwargs):
        """
        Convert :math:`\\frac{d \\sigma}{d \\Omega}(Q)` to :math:`S(Q)`

        :param q: :math:`Q`-space vector
        :type q: numpy.array or list
        :param dcs: :math:`\\frac{d \\sigma}{d \\Omega}(Q)` vector
        :type dcs: numpy.array or list
        :param ddcs: uncertainty vector
        :type ddcs: numpy.array or list

        :return: (:math:`S(Q)` vector, uncertainty vector)
        :rtype: (numpy.array, numpy.array)
        """
        fq, dfq = self.DCS_to_FK(q, dcs, ddcs, **kwargs)
        return self.FK_to_S(q, fq, dfq, **kwargs)

    def DCS_to_FK(self, q, dcs, ddcs=None, **kwargs):
        """
        Convert :math:`\\frac{d \\sigma}{d \\Omega}(Q)` to :math:`F(Q)`

        :param q: :math:`Q`-space vector
        :type q: numpy.array or list
        :param fq: :math:`\\frac{d \\sigma}{d \\Omega}(Q)` vector
        :type fq: numpy.array or list
        :param ddcs: uncertainty vector
        :type ddcs: numpy.array or list

        :return: (:math:`F(Q)` vector, uncertainty vector)
        :rtype: (numpy.array, numpy.array)
        """
        return (dcs - kwargs['<b_tot^2>'], ddcs)

    # Real Space Conversions

    # G(r) = PDF
    def G_to_GK(self, r, gr, dgr=None, **kwargs):
        r"""
        Convert :math:`G_{PDFFIT}(r)` to :math:`G_{Keen Version}(r)`

        :param r: r-space vector
        :type r: numpy.array or list
        :param gr: :math:`G_{PDFFIT}(r)` vector
        :type gr: numpy.array or list
        :param dgr: uncertainty vector :math:`\Delta g(r)`
        :type dgr: numpy.array or list

        :return: :math:`G_{Keen Version}(r)` vector, uncertainty vector
        :rtype: (numpy.array, numpy.array)
        """
        factor = kwargs['<b_coh>^2'] / (4. * np.pi * kwargs['rho'])
        if dgr is None:
            dgr = np.zeros_like(gr)
        return (factor * self._safe_divide(gr, r),
                factor * self._safe_divide(dgr, r))

    def G_to_g(self, r, gr, dgr=None, **kwargs):
        r"""
        Convert :math:`G_{PDFFIT}(r)` to :math:`g(r)`

        :param r: r-space vector
        :type r: numpy.array or list
        :param gr: :math:`G_{PDFFIT}(r)` vector
        :type gr: numpy.array or list
        :param dgr: uncertainty vector :math:`\Delta g(r)`
        :type dgr: numpy.array or list

        :return: :math:`g(r)` vector, uncertainty vector
        :rtype: (numpy.array, numpy.array)
        """
        factor = 4. * np.pi * kwargs['rho']
        if dgr is None:
            dgr = np.zeros_like(gr)
        return (self._safe_divide(gr, factor * r) + 1.,
                self._safe_divide(dgr, factor * r))

    # Keen's G(r)
    def GK_to_G(self, r, gr, dgr=None, **kwargs):
        r"""
        Convert :math:`G_{Keen Version}(r)` to :math:`G_{PDFFIT}(r)`

        :param r: r-space vector
        :type r: numpy.array or list
        :param gr: :math:`G_{Keen Version}(r)` vector
        :type gr: numpy.array or list
        :param dgr: uncertainty vector :math:`\Delta g(r)`
        :type dgr: numpy.array or list

        :return: :math:`G_{PDFFIT}(r)` vector, uncertainty vector
        :rtype: (numpy.array, numpy.array)
        """
        factor = (4. * np.pi * kwargs['rho']) / kwargs['<b_coh>^2']
        if dgr is None:
            dgr = np.zeros_like(gr)
        return (factor * r * gr, factor * r * dgr)

    def GK_to_g(self, r, gr, dgr=None, **kwargs):
        r"""
        Convert :math:`G_{Keen Version}(r)` to :math:`g(r)`

        :param r: r-space vector
        :type r: numpy.array or list
        :param gr: :math:`G_{Keen Version}(r)` vector
        :type gr: numpy.array or list
        :param dgr: uncertainty vector :math:`\Delta g(r)`
        :type dgr: numpy.array or list

        :return: :math:`g(r)` vector, uncertainty vector
        :rtype: (numpy.array, numpy.array)
        """
        _gr, _dgr = self.GK_to_G(r, gr, dgr=dgr, **kwargs)
        return self.G_to_g(r, _gr, dgr=_dgr, **kwargs)

    # g(r)
    def g_to_G(self, r, gr, dgr=None, **kwargs):
        r"""
        Convert :math:`g(r)` to :math:`G_{PDFFIT}(r)`

        :param r: r-space vector
        :type r: numpy.array or list
        :param gr: :math:`g(r)` vector
        :type gr: numpy.array or list
        :param dgr: uncertainty vector :math:`\Delta g(r)`
        :type dgr: numpy.array or list

        :return: :math:`G_{PDFFIT}(r)` vector, uncertainty vector
        :rtype: (numpy.array, numpy.array)
        """
        factor = 4. * np.pi * r * kwargs['rho']
        if dgr is None:
            dgr = np.zeros_like(gr)
        return (factor * (gr - 1.), factor * dgr)

    def g_to_GK(self, r, gr, dgr=None, **kwargs):
        r"""
        Convert :math:`g(r)` to :math:`G_{Keen Version}(r)`

        :param r: r-space vector
        :type r: numpy.array or list
        :param gr: :math:`g(r)` vector
        :type gr: numpy.array or list
        :param dgr: uncertainty vector :math:`\Delta g(r)`
        :type dgr: numpy.array or list

        :return: :math:`G_{Keen Version}(r)` vector, uncertainty vector
        :rtype: (numpy.array, numpy.array)
        """
        _gr, _dgr = self.g_to_G(r, gr, dgr=dgr, **kwargs)
        return self.G_to_GK(r, _gr, dgr=_dgr, **kwargs)



import unittest
import numpy as np
from numpy.testing import assert_allclose, assert_array_equal
from tests.utils import \
    load_data, get_index_of_function
from tests.materials import Nickel, Argon
from pystog.utils import \
    RealSpaceHeaders, ReciprocalSpaceHeaders
from pystog.converter import Converter


class TestConverterUtilities(unittest.TestCase):
    def setUp(self):
        unittest.TestCase.setUp(self)
        self.converter = Converter()

    def tearDown(self):
        unittest.TestCase.tearDown(self)

    def test_safe_divide(self):
        assert_array_equal(
            self.converter._safe_divide(np.arange(10), np.arange(10)),
            [0, 1, 1, 1, 1, 1, 1, 1, 1, 1])


# Real Space Function


class TestConverterRealSpaceBase(unittest.TestCase):
    rtol = 1e-5
    atol = 1e-8

    def initialize_material(self):
        # setup input data
        self.kwargs = self.material.kwargs

        # setup the tolerance
        self.first = self.material.real_space_first
        self.last = self.material.real_space_last

        data = load_data(self.material.real_space_filename)
        self.r = data[:, get_index_of_function("r", RealSpaceHeaders)]
        self.gofr = data[:, get_index_of_function("g(r)", RealSpaceHeaders)]
        self.GofR = data[:, get_index_of_function("G(r)", RealSpaceHeaders)]
        self.GKofR = data[:, get_index_of_function("GK(r)", RealSpaceHeaders)]

        # targets for 1st peaks
        self.gofr_target = self.material.gofr_target
        self.GofR_target = self.material.GofR_target
        self.GKofR_target = self.material.GKofR_target

    def setUp(self):
        unittest.TestCase.setUp(self)
        self.converter = Converter()

    def tearDown(self):
        unittest.TestCase.tearDown(self)

    # g(r) tests
    def g_to_G(self):
        GofR, dGofR = self.converter.g_to_G(self.r, self.gofr,
                                            np.ones_like(self.r),
                                            **self.kwargs)

        FourPiRhoR = 4 * np.pi * self.kwargs['rho'] * self.r

        assert_allclose(
            GofR[self.first:self.last],
            self.GofR_target,
            rtol=self.rtol,
            atol=self.atol)
        assert_allclose(dGofR, np.ones_like(self.r) * FourPiRhoR)

    def g_to_GK(self):
        GKofR, dGKofR = self.converter.g_to_GK(self.r, self.gofr,
                                               np.ones_like(self.r),
                                               **self.kwargs)
        assert_allclose(
            GKofR[self.first:self.last],
            self.GKofR_target,
            rtol=self.rtol,
            atol=self.atol)
        assert_allclose(dGKofR,
                        np.ones_like(self.r) * self.kwargs['<b_coh>^2'])

    # G(r) tests
    def G_to_g(self):
        gofr, dgofr = self.converter.G_to_g(self.r, self.GofR,
                                            np.ones_like(self.r),
                                            **self.kwargs)

        FourPiRhoR = 4 * np.pi * self.kwargs['rho'] * self.r

        assert_allclose(
            gofr[self.first:self.last],
            self.gofr_target,
            rtol=self.rtol,
            atol=self.atol)
        assert_allclose(dgofr, np.ones_like(self.r) / FourPiRhoR)

    def G_to_GK(self):
        GKofR, dGKofR = self.converter.G_to_GK(self.r, self.GofR,
                                               np.ones_like(self.r),
                                               **self.kwargs)
        FourPiRhoR = 4 * np.pi * self.kwargs['rho'] * self.r
        bcoh_sqrd = self.kwargs['<b_coh>^2']

        assert_allclose(
            GKofR[self.first:self.last],
            self.GKofR_target,
            rtol=self.rtol,
            atol=self.atol)
        assert_allclose(dGKofR,
                        np.ones_like(self.r) * bcoh_sqrd / FourPiRhoR)

    # GK(r) tests
    def GK_to_g(self):
        gofr, dgofr = self.converter.GK_to_g(self.r, self.GKofR,
                                             np.ones_like(self.r),
                                             **self.kwargs)
        assert_allclose(
            gofr[self.first:self.last],
            self.gofr_target,
            rtol=self.rtol,
            atol=self.atol)
        assert_allclose(dgofr, np.ones_like(self.r) / self.kwargs['<b_coh>^2'])

    def GK_to_G(self):
        GofR, dGofR = self.converter.GK_to_G(self.r, self.GKofR,
                                             np.ones_like(self.r),
                                             **self.kwargs)

        FourPiRhoR = 4 * np.pi * self.kwargs['rho'] * self.r
        bcoh_sqrd = self.kwargs['<b_coh>^2']

        assert_allclose(
            GofR[self.first:self.last],
            self.GofR_target,
            rtol=self.rtol,
            atol=self.atol)
        assert_allclose(dGofR,
                        np.ones_like(self.r) / bcoh_sqrd * FourPiRhoR)


class TestConverterRealSpaceNickel(TestConverterRealSpaceBase):
    def setUp(self):
        super(TestConverterRealSpaceNickel, self).setUp()
        self.material = Nickel()
        self.initialize_material()

    def test_g_to_G(self):
        self.g_to_G()

    def test_g_to_GK(self):
        self.g_to_GK()

    def test_G_to_g(self):
        self.G_to_g()

    def test_G_to_GK(self):
        self.G_to_GK()

    def test_GK_to_g(self):
        self.GK_to_g()

    def test_GK_to_G(self):
        self.GK_to_G()


class TestConverterRealSpaceArgon(TestConverterRealSpaceBase):
    def setUp(self):
        super(TestConverterRealSpaceArgon, self).setUp()
        self.material = Argon()
        self.initialize_material()

    def test_g_to_G(self):
        self.g_to_G()

    def test_g_to_GK(self):
        self.g_to_GK()

    def test_G_to_g(self):
        self.G_to_g()

    def test_G_to_GK(self):
        self.G_to_GK()

    def test_GK_to_g(self):
        self.GK_to_g()

    def test_GK_to_G(self):
        self.GK_to_G()


# Reciprocal Space Function


class TestConverterReciprocalSpaceBase(unittest.TestCase):
    rtol = 1e-5
    atol = 1e-8

    def initialize_material(self):
        # setup input data
        self.kwargs = self.material.kwargs

        # setup the tolerance
        self.first = self.material.reciprocal_space_first
        self.last = self.material.reciprocal_space_last

        data = load_data(self.material.reciprocal_space_filename)
        self.q = data[:, get_index_of_function("Q", ReciprocalSpaceHeaders)]
        self.sq = data[:,
                       get_index_of_function("S(Q)", ReciprocalSpaceHeaders)]
        self.fq = data[:,
                       get_index_of_function("Q[S(Q)-1]",
                                             ReciprocalSpaceHeaders)]
        self.fq_keen = data[:,
                            get_index_of_function("FK(Q)",
                                                  ReciprocalSpaceHeaders)]
        self.dcs = data[:,
                        get_index_of_function("DCS(Q)",
                                              ReciprocalSpaceHeaders)]

        # targets for 1st peaks
        self.sq_target = self.material.sq_target
        self.fq_target = self.material.fq_target
        self.fq_keen_target = self.material.fq_keen_target
        self.dcs_target = self.material.dcs_target

    def setUp(self):
        unittest.TestCase.setUp(self)
        self.converter = Converter()

    def tearDown(self):
        unittest.TestCase.tearDown(self)

    # S(Q) tests
    def S_to_F(self):
        fq, dfq = self.converter.S_to_F(self.q, self.sq, np.ones_like(self.q),
                                        **self.kwargs)
        assert_allclose(
            fq[self.first:self.last],
            self.fq_target,
            rtol=self.rtol,
            atol=self.atol)
        assert_allclose(dfq, np.ones_like(self.q) * self.q)

    def S_to_FK(self):
        fq_keen, dfq_keen = self.converter.S_to_FK(self.q, self.sq,
                                                   np.ones_like(self.q),
                                                   **self.kwargs)
        assert_allclose(
            fq_keen[self.first:self.last],
            self.fq_keen_target,
            rtol=self.rtol,
            atol=self.atol)
        assert_allclose(dfq_keen,
                        np.ones_like(self.q) * self.kwargs['<b_coh>^2'])

    def S_to_DCS(self):
        dcs, ddcs = self.converter.S_to_DCS(self.q, self.sq,
                                            np.ones_like(self.q),
                                            **self.kwargs)
        assert_allclose(
            dcs[self.first:self.last],
            self.dcs_target,
            rtol=self.rtol,
            atol=self.atol)
        assert_allclose(ddcs, np.ones_like(self.q) * self.kwargs['<b_coh>^2'])

    # Q[S(Q)-1] tests

    def F_to_S(self):
        sq, dsq = self.converter.F_to_S(self.q, self.fq, np.ones_like(self.q),
                                        **self.kwargs)
        assert_allclose(
            sq[self.first:self.last],
            self.sq_target,
            rtol=self.rtol,
            atol=self.atol)
        assert_allclose(dsq, np.ones_like(self.q) / self.q)

    def F_to_S_with_no_dfq(self):
        sq, dsq = self.converter.F_to_S(self.q, self.fq, **self.kwargs)
        assert_allclose(
            sq[self.first:self.last],
            self.sq_target,
            rtol=self.rtol,
            atol=self.atol)
        assert_allclose(dsq, np.zeros_like(dsq))

    def F_to_FK(self):
        fq_keen, dfq_keen = self.converter.F_to_FK(
            self.q, self.fq, np.ones_like(self.q), **self.kwargs)
        assert_allclose(
            fq_keen[self.first:self.last],
            self.fq_keen_target,
            rtol=self.rtol,
            atol=self.atol)
        assert_allclose(
            dfq_keen,
            np.ones_like(self.q) * self.kwargs['<b_coh>^2'] / self.q)

    def F_to_FK_with_no_dfq(self):
        fq_keen, dfq_keen = self.converter.F_to_FK(
            self.q, self.fq, **self.kwargs)
        assert_allclose(
            fq_keen[self.first:self.last],
            self.fq_keen_target,
            rtol=self.rtol,
            atol=self.atol)
        assert_allclose(dfq_keen, np.zeros_like(dfq_keen))

    def F_to_DCS(self):
        dcs, ddcs = self.converter.F_to_DCS(self.q, self.fq,
                                            np.ones_like(self.q),
                                            **self.kwargs)
        assert_allclose(
            dcs[self.first:self.last],
            self.dcs_target,
            rtol=self.rtol,
            atol=self.atol)
        assert_allclose(
            ddcs,
            np.ones_like(self.q) * self.kwargs['<b_coh>^2'] / self.q)

    # FK(Q) tests

    def FK_to_S(self):
        sq, dsq = self.converter.FK_to_S(self.q, self.fq_keen,
                                         np.ones_like(self.q), **self.kwargs)
        assert_allclose(
            sq[self.first:self.last],
            self.sq_target,
            rtol=self.rtol,
            atol=self.atol)
        assert_allclose(dsq, np.ones_like(self.q) / self.kwargs['<b_coh>^2'])

    def FK_to_F(self):
        fq, dfq = self.converter.FK_to_F(self.q, self.fq_keen,
                                         np.ones_like(self.q), **self.kwargs)
        assert_allclose(
            fq[self.first:self.last],
            self.fq_target,
            rtol=self.rtol,
            atol=self.atol)
        assert_allclose(
            dfq,
            np.ones_like(self.q) / self.kwargs['<b_coh>^2'] * self.q)

    def FK_to_DCS(self):
        dcs, ddcs = self.converter.FK_to_DCS(self.q, self.fq_keen,
                                             np.ones_like(self.q),
                                             **self.kwargs)
        assert_allclose(
            dcs[self.first:self.last],
            self.dcs_target,
            rtol=self.rtol,
            atol=self.atol)
        assert_allclose(ddcs, np.ones_like(self.q))

    # DCS(Q) tests

    def DCS_to_S(self):
        sq, dsq = self.converter.DCS_to_S(self.q, self.dcs, np.ones_like(
            self.q), **self.kwargs)
        assert_allclose(
            sq[self.first:self.last],
            self.sq_target,
            rtol=self.rtol,
            atol=self.atol)
        assert_allclose(dsq, np.ones_like(self.q) / self.kwargs['<b_coh>^2'])

    def DCS_to_F(self):
        fq, dfq = self.converter.DCS_to_F(self.q, self.dcs, np.ones_like(
            self.q), **self.kwargs)
        assert_allclose(
            fq[self.first:self.last],
            self.fq_target,
            rtol=self.rtol,
            atol=self.atol)
        assert_allclose(
            dfq,
            np.ones_like(self.q) / self.kwargs['<b_coh>^2'] * self.q)

    def DCS_to_FK(self):
        fq_keen, dfq_keen = self.converter.DCS_to_FK(self.q, self.dcs,
                                                     np.ones_like(self.q),
                                                     **self.kwargs)
        assert_allclose(
            fq_keen[self.first:self.last],
            self.fq_keen_target,
            rtol=self.rtol,
            atol=self.atol)
        assert_allclose(dfq_keen, np.ones_like(self.q))


class TestConverterReciprocalSpaceNickel(TestConverterReciprocalSpaceBase):
    def setUp(self):
        super(TestConverterReciprocalSpaceNickel, self).setUp()
        self.material = Nickel()
        self.initialize_material()

    def test_S_to_F(self):
        self.S_to_F()

    def test_S_to_FK(self):
        self.S_to_FK()

    def test_S_to_DCS(self):
        self.S_to_DCS()

    def test_F_to_S(self):
        self.F_to_S()

    def test_F_to_S_with_no_dfq(self):
        self.F_to_S_with_no_dfq()

    def test_F_to_FK(self):
        self.F_to_FK()

    def test_F_to_FK_with_no_dfq(self):
        self.F_to_FK_with_no_dfq()

    def test_F_to_DCS(self):
        self.F_to_DCS()

    def test_FK_to_S(self):
        self.FK_to_S()

    def test_FK_to_F(self):
        self.FK_to_F()

    def test_FK_to_DCS(self):
        self.FK_to_DCS()

    def test_DCS_to_S(self):
        self.DCS_to_S()

    def test_DCS_to_F(self):
        self.DCS_to_F()

    def test_DCS_to_FK(self):
        self.DCS_to_FK()


class TestConverterReciprocalSpaceArgon(TestConverterReciprocalSpaceBase):
    def setUp(self):
        super(TestConverterReciprocalSpaceArgon, self).setUp()
        self.material = Argon()
        self.initialize_material()

    def test_S_to_F(self):
        self.S_to_F()

    def test_S_to_FK(self):
        self.S_to_FK()

    def test_S_to_DCS(self):
        self.S_to_DCS()

    def test_F_to_S(self):
        self.F_to_S()

    def test_F_to_S_with_no_dfq(self):
        self.F_to_S_with_no_dfq()

    def test_F_to_FK(self):
        self.F_to_FK()

    def test_F_to_FK_with_no_dfq(self):
        self.F_to_FK_with_no_dfq()

    def test_F_to_DCS(self):
        self.F_to_DCS()

    def test_FK_to_S(self):
        self.FK_to_S()

    def test_FK_to_F(self):
        self.FK_to_F()

    def test_FK_to_DCS(self):
        self.FK_to_DCS()

    def test_DCS_to_S(self):
        self.DCS_to_S()

    def test_DCS_to_F(self):
        self.DCS_to_F()

    def test_DCS_to_FK(self):
        self.DCS_to_FK()


if __name__ == '__main__':
    unittest.main()  # pragma: no cover



import pytest
from pystog.cli import pystog_cli
from pystog.stog import NoInputFilesException


def test_pystog_cli_no_files_exception():
    with pytest.raises(NoInputFilesException):
        pystog_cli({"cat": "meow"})



"""
==============
FourierFilter
==============

This module defines the FourierFilter class
that performs the Fourier filter for a
given range to exclude.
"""

import numpy as np

from pystog.converter import Converter
from pystog.transformer import Transformer


class FourierFilter:
    """
    The FourierFilter class is used to exlude a given
    range in the current function by a back Fourier Transform
    of that section, followed by a difference from the non-excluded
    function, and then a forward transform of the difference function
    Can currently do:
    a real space function -> reciprocal space function -> real space function

    :examples:

    >>> import numpy
    >>> from pystog import FourierFilter
    >>> ff = FourierFilter()
    >>> r, gr = numpy.loadtxt("my_gofr_file.txt",unpack=True)
    >>> q = numpy.linspace(0., 25., 2500)
    >>> q, sq = transformer.G_to_S(r, gr, q)
    >>> q_ft, sq_ft, q, sq, r, gr = ff.G_using_F(r, gr, q, sq, 1.5)
    """

    def __init__(self):
        self.converter = Converter()
        self.transformer = Transformer()

    # g(r)
    def g_using_F(self, r, gr, q, fq, cutoff, dgr=None, dfq=None, **kwargs):
        """
        Fourier filters real space :math:`g(r)`
        using the reciprocal space :math:`Q[S(Q)-1]`

        :param r: :math:`r`-space vector
        :type r: numpy.array or list
        :param gr: :math:`g(r)` vector
        :type gr: numpy.array or list
        :param q: :math:`Q`-space vector
        :type q: numpy.array or list
        :param fq: :math:`Q[S(Q)-1]` vector
        :type fq: numpy.array or list
        :param cutoff: The :math:`r_{max}` value to filter from 0. to cutoff
        :type cutoff: float

        :return: A tuple of the :math:`Q` and :math:`Q[S(Q)-1]`
                 for the 0. to cutoff transform,
                 the corrected :math:`Q` and :math:`Q[S(Q)-1]`,
                 and the filtered :math:`r` and :math:`g(r)`.

                 Thus,
                 [
                    :math:`Q_{FF}`,
                    :math:`Q[S(Q)-1]_{FF}`,
                    :math:`Q`,
                    :math:`Q[S(Q)-1]`,
                    :math:`r_{FF}`,
                    :math:`g(r)_{FF}`
                ]
        :rtype: tuple of numpy.array
        """
        # setup qmin, qmax, and get low-r region to back transform
        qmin = min(q)
        qmax = max(q)
        r_tmp, gr_tmp_init, dgr_tmp_init = self.transformer.apply_cropping(
            r, gr, 0.0, cutoff, dy=dgr)

        # Shift low-r so it goes to 1 at "high-r" for this section. Reduces the
        # sinc function issue.
        gr_tmp = gr_tmp_init + 1

        # Transform the shifted low-r region to F(Q) to get F(Q)_ft
        q_ft, fq_ft, dfq_ft = self.transformer.g_to_F(
            r_tmp, gr_tmp, q, dgr=dgr_tmp_init, **kwargs)

        q_ft, fq_ft, dfq_ft = self.transformer.apply_cropping(
            q_ft, fq_ft, qmin, qmax, dy=dfq_ft)

        # Subtract F(Q)_ft from original F(Q) = delta_F(Q)
        q, fq, dfq = self.transformer.apply_cropping(
            q, fq, qmin, qmax, dy=dfq)
        fq = (fq - fq_ft)
        dfq = np.sqrt(dfq**2 + dfq_ft**2)

        # Transform delta_F(Q) for g(r) with low-r removed
        r, gr, dgr = self.transformer.F_to_g(q, fq, r, dfq=dfq, **kwargs)

        return q_ft, fq_ft, q, fq, r, gr, dfq_ft, dfq, dgr

    def g_using_S(self, r, gr, q, sq, cutoff, dgr=None, dsq=None, **kwargs):
        """
        Fourier filters real space :math:`g(r)`
        using the reciprocal space :math:`S(Q)`

        :param r: :math:`r`-space vector
        :type r: numpy.array or list
        :param gr: :math:`g(r)` vector
        :type gr: numpy.array or list
        :param q: :math:`Q`-space vector
        :type q: numpy.array or list
        :param sq: :math:`S(Q)` vector
        :type sq: numpy.array or list
        :param cutoff: The :math:`r_{max}` value to filter from 0. to cutoff
        :type cutoff: float

        :return: A tuple of the :math:`Q` and :math:`S(Q)`
                 for the 0. to cutoff transform,
                 the corrected :math:`Q` and :math:`S(Q)`,
                 and the filtered :math:`r` and :math:`g(r)`.

                 Thus,
                 [:math:`Q_{FF}`, :math:`S(Q)_{FF}`,
                 :math:`Q`, :math:`S(Q)`, :math:`r_{FF}`, :math:`g(r)_{FF}]`
        :rtype: tuple of numpy.array
        """
        fq, dfq = self.converter.S_to_F(q, sq, dsq=dsq)
        q_ft, fq_ft, q, fq, r, gr, dfq_ft, dfq, dgr = self.g_using_F(
            r, gr, q, fq, cutoff, dgr=dgr, dfq=dfq, **kwargs)
        sq_ft, dsq_ft = self.converter.F_to_S(q_ft, fq_ft, dfq=dfq_ft)
        sq, dsq = self.converter.F_to_S(q, fq, dfq=dfq)
        return q_ft, sq_ft, q, sq, r, gr, dsq_ft, dsq, dgr

    def g_using_FK(self, r, gr, q, fq, cutoff, dgr=None, dfq=None, **kwargs):
        """
        Fourier filters real space :math:`g(r)`
        using the reciprocal space :math:`F(Q)`

        :param r: :math:`r`-space vector
        :type r: numpy.array or list
        :param gr: :math:`g(r)` vector
        :type gr: numpy.array or list
        :param q: :math:`Q`-space vector
        :type q: numpy.array or list
        :param fq: :math:`F(Q)` vector
        :type fq: numpy.array or list
        :param cutoff: The :math:`r_{max}` value to filter from 0. to cutoff
        :type cutoff: float

        :return: A tuple of the :math:`Q` and :math:`F(Q)`
                 for the 0. to cutoff transform,
                 the corrected :math:`Q` and :math:`F(Q)`,
                 and the filtered :math:`r` and :math:`g(r)`.

                 Thus,
                 [:math:`Q_{FF}`, :math:`F(Q)_{FF}`,
                 :math:`Q`, :math:`F(Q)`, :math:`r_{FF}`, :math:`g(r)_{FF}]`
        :rtype: tuple of numpy.array
        """
        fq, dfq = self.converter.FK_to_F(q, fq, dfq=dfq, **kwargs)
        q_ft, fq_ft, q, fq, r, gr, dfq_ft, dfq, dgr = self.g_using_F(
            r, gr, q, fq, cutoff, dgr=dgr, dfq=dfq, **kwargs)
        fq_ft, dfq_ft = self.converter.F_to_FK(
            q_ft, fq_ft, dfq=dfq_ft, **kwargs)
        fq, dfq = self.converter.F_to_FK(q, fq, dfq=dfq, **kwargs)
        return q_ft, fq_ft, q, fq, r, gr, dfq_ft, dfq, dgr

    def g_using_DCS(self, r, gr, q, dcs, cutoff, dgr=None, ddcs=None, **kwargs):
        """
        Fourier filters real space :math:`g(r)`
        using the reciprocal space
        :math:`\\frac{d \\sigma}{d \\Omega}(Q)`

        :param r: :math:`r`-space vector
        :type r: numpy.array or list
        :param gr: :math:`g(r)` vector
        :type gr: numpy.array or list
        :param q: :math:`Q`-space vector
        :type q: numpy.array or list
        :param dcs: :math:`\\frac{d \\sigma}{d \\Omega}(Q)` vector
        :type dcs: numpy.array or list
        :param cutoff: The :math:`r_{max}` value to filter from 0. to cutoff
        :type cutoff: float

        :return: A tuple of the :math:`Q` and
                 :math:`\\frac{d \\sigma}{d \\Omega}(Q)`
                 for the 0. to cutoff transform,
                 the corrected :math:`Q` and
                 :math:`\\frac{d \\sigma}{d \\Omega}(Q)`,
                 and the filtered :math:`r` and :math:`g(r)`.

                 Thus,
                 [:math:`Q_{FF}`, :math:`\\frac{d \\sigma}{d \\Omega}(Q)_{FF}`,
                 :math:`Q`, :math:`\\frac{d \\sigma}{d \\Omega}(Q)`,
                 :math:`r_{FF}`, :math:`g(r)_{FF}]`
        :rtype: tuple of numpy.array
        """
        fq, dfq = self.converter.DCS_to_F(q, dcs, ddcs=ddcs, **kwargs)
        q_ft, fq_ft, q, fq, r, gr, dfq_ft, dfq, dgr = self.g_using_F(
            r, gr, q, fq, cutoff, dgr=dgr, dfq=dfq, **kwargs)
        dcs_ft, ddcs_ft = self.converter.F_to_DCS(
            q_ft, fq_ft, dfq=dfq_ft, **kwargs)
        dcs, ddcs = self.converter.F_to_DCS(q_ft, fq, dfq=dfq, **kwargs)
        return q_ft, dcs_ft, q, dcs, r, gr, ddcs_ft, ddcs, dgr

    # G(R) = PDF
    def G_using_F(self, r, gr, q, fq, cutoff, dgr=None, dfq=None, **kwargs):
        """
        Fourier filters real space :math:`G_{PDFFIT}(r)`
        using the reciprocal space :math:`Q[S(Q)-1]`

        :param r: :math:`r`-space vector
        :type r: numpy.array or list
        :param gr: :math:`G_{PDFFIT}(r)` vector
        :type gr: numpy.array or list
        :param q: :math:`Q`-space vector
        :type q: numpy.array or list
        :param fq: :math:`Q[S(Q)-1]` vector
        :type fq: numpy.array or list
        :param cutoff: The :math:`r_{max}` value to filter from 0. to cutoff
        :type cutoff: float

        :return: A tuple of the :math:`Q` and :math:`Q[S(Q)-1]`
                 for the 0. to cutoff transform,
                 the corrected :math:`Q` and :math:`Q[S(Q)-1]`,
                 and the filtered :math:`r` and :math:`G_{PDFFIT}(r)`.

                 Thus,
                 [
                    :math:`Q_{FF}`,
                    :math:`Q[S(Q)-1]_{FF}`,
                    :math:`Q`,
                    :math:`Q[S(Q)-1]`,
                    :math:`r_{FF}`,
                    :math:`G_{PDFFIT}(r)_{FF}`
                ]
        :rtype: tuple of numpy.array
        """
        gr, dgr = self.converter.G_to_g(r, gr, dgr=dgr, **kwargs)
        q_ft, fq_ft, q, fq, r, gr, dfq_ft, dfq, dgr = self.g_using_F(
            r, gr, q, fq, cutoff, dgr=dgr, dfq=dfq, **kwargs)
        gr, dgr = self.converter.g_to_G(r, gr, dgr=dgr, **kwargs)
        return q_ft, fq_ft, q, fq, r, gr, dfq_ft, dfq, dgr

    def G_using_S(self, r, gr, q, sq, cutoff, dgr=None, dsq=None, **kwargs):
        """
        Fourier filters real space :math:`G_{PDFFIT}(r)`
        using the reciprocal space :math:`S(Q)`

        :param r: :math:`r`-space vector
        :type r: numpy.array or list
        :param gr: :math:`G_{PDFFIT}(r)` vector
        :type gr: numpy.array or list
        :param q: :math:`Q`-space vector
        :type q: numpy.array or list
        :param fq: :math:`S(Q)` vector
        :type fq: numpy.array or list
        :param cutoff: The :math:`r_{max}` value to filter from 0. to cutoff
        :type cutoff: float

        :return: A tuple of the :math:`Q` and :math:`S(Q)`
                 for the 0. to cutoff transform,
                 the corrected :math:`Q` and :math:`S(Q)`,
                 and the filtered :math:`r` and :math:`G_{PDFFIT}(r)`.

                 Thus,
                 [
                    :math:`Q_{FF}`,
                    :math:`S(Q)_{FF}`,
                    :math:`Q`,
                    :math:`S(Q)`,
                    :math:`r_{FF}`,
                    :math:`G_{PDFFIT}(r)_{FF}`
                ]
        :rtype: tuple of numpy.array
        """
        fq, dfq = self.converter.S_to_F(q, sq, dsq=dsq)
        q_ft, fq_ft, q, fq, r, gr, dfq_ft, dfq, dgr = self.G_using_F(
            r, gr, q, fq, cutoff, dgr=dgr, dfq=dfq, **kwargs)
        sq_ft, dsq_ft = self.converter.F_to_S(q_ft, fq_ft, dfq_ft)
        sq, dsq = self.converter.F_to_S(q, fq, dfq)
        return q_ft, sq_ft, q, sq, r, gr, dsq_ft, dsq, dgr

    def G_using_FK(self, r, gr, q, fq, cutoff, dgr=None, dfq=None, **kwargs):
        """
        Fourier filters real space :math:`G_{PDFFIT}(r)`
        using the reciprocal space :math:`F(Q)`

        :param r: :math:`r`-space vector
        :type r: numpy.array or list
        :param gr: :math:`G_{PDFFIT}(r)` vector
        :type gr: numpy.array or list
        :param q: :math:`Q`-space vector
        :type q: numpy.array or list
        :param fq: :math:`F(Q)` vector
        :type fq: numpy.array or list
        :param cutoff: The :math:`r_{max}` value to filter from 0. to cutoff
        :type cutoff: float

        :return: A tuple of the :math:`Q` and :math:`F(Q)`
                 for the 0. to cutoff transform,
                 the corrected :math:`Q` and :math:`F(Q)`,
                 and the filtered :math:`r` and :math:`G_{PDFFIT}(r)`.

                 Thus,
                 [
                    :math:`Q_{FF}`,
                    :math:`F(Q)_{FF}`,
                    :math:`Q`,
                    :math:`F(Q)`,
                    :math:`r_{FF}`,
                    :math:`G_{PDFFIT}(r)_{FF}`
                ]
        :rtype: tuple of numpy.array
        """
        fq, dfq = self.converter.FK_to_F(q, fq, dfq=dfq, **kwargs)
        q_ft, fq_ft, q, fq, r, gr, dfq_ft, dfq, dgr = self.G_using_F(
            r, gr, q, fq, cutoff, dgr=dgr, dfq=dfq, **kwargs)
        fq_ft, dfq_ft = self.converter.F_to_FK(
            q_ft, fq_ft, dfq=dfq_ft, **kwargs)
        fq, dfq = self.converter.F_to_FK(q, fq, dfq=dfq, **kwargs)
        return q_ft, fq_ft, q, fq, r, gr, dfq_ft, dfq, dgr

    def G_using_DCS(self, r, gr, q, dcs, cutoff, dgr=None, ddcs=None, **kwargs):
        """
        Fourier filters real space :math:`G_{PDFFIT}(r)`
        using the reciprocal space
        :math:`\\frac{d \\sigma}{d \\Omega}(Q)`

        :param r: :math:`r`-space vector
        :type r: numpy.array or list
        :param gr: :math:`G_{PDFFIT}(r)` vector
        :type gr: numpy.array or list
        :param q: :math:`Q`-space vector
        :type q: numpy.array or list
        :param dcs: :math:`\\frac{d \\sigma}{d \\Omega}(Q)` vector
        :type dcs: numpy.array or list
        :param cutoff: The :math:`r_{max}` value to filter from 0. to cutoff
        :type cutoff: float

        :return: A tuple of the :math:`Q` and
                 :math:`\\frac{d \\sigma}{d \\Omega}(Q)`
                 for the 0. to cutoff transform,
                 the corrected :math:`Q` and
                 :math:`\\frac{d \\sigma}{d \\Omega}(Q)`,
                 and the filtered :math:`r` and :math:`G_{PDFFIT}(r)`.

                 Thus,
                 [:math:`Q_{FF}`, :math:`\\frac{d \\sigma}{d \\Omega}(Q)_{FF}`,
                 :math:`Q`, :math:`\\frac{d \\sigma}{d \\Omega}(Q)`,
                 :math:`r_{FF}`, :math:`G_{PDFFIT}(r)_{FF}]`
        :rtype: tuple of numpy.array
        """
        fq, dfq = self.converter.DCS_to_F(q, dcs, ddcs=ddcs, **kwargs)
        q_ft, fq_ft, q, fq, r, gr, dfq_ft, dfq, dgr = self.G_using_F(
            r, gr, q, fq, cutoff, dgr=dgr, dfq=dfq, **kwargs)
        dcs_ft, ddcs_ft = self.converter.F_to_DCS(
            q_ft, fq_ft, dfq=dfq_ft, **kwargs)
        dcs, ddcs = self.converter.F_to_DCS(q, fq, dfq=dfq, **kwargs)
        return q_ft, dcs_ft, q, dcs, r, gr, ddcs_ft, ddcs, dgr

    # Keen's G(r)
    def GK_using_F(self, r, gr, q, fq, cutoff, dgr=None, dfq=None, **kwargs):
        """
        Fourier filters real space :math:`G_{Keen Version}(r)`
        using the reciprocal space :math:`Q[S(Q)-1]`

        :param r: :math:`r`-space vector
        :type r: numpy.array or list
        :param gr: :math:`G_{Keen Version}(r)` vector
        :type gr: numpy.array or list
        :param q: :math:`Q`-space vector
        :type q: numpy.array or list
        :param fq: :math:`Q[S(Q)-1]` vector
        :type fq: numpy.array or list
        :param cutoff: The :math:`r_{max}` value to filter from 0. to cutoff
        :type cutoff: float

        :return: A tuple of the :math:`Q` and :math:`Q[S(Q)-1]`
                 for the 0. to cutoff transform,
                 the corrected :math:`Q` and :math:`Q[S(Q)-1]`,
                 and the filtered :math:`r` and :math:`G_{Keen Version}(r)`.

                 Thus,
                 [
                    :math:`Q_{FF}`,
                    :math:`Q[S(Q)-1]_{FF}`,
                    :math:`Q`,
                    :math:`Q[S(Q)-1]`,
                    :math:`r_{FF}`,
                    :math:`G_{Keen Version}(r)_{FF}`
                ]
        :rtype: tuple of numpy.array
        """

        gr, dgr = self.converter.GK_to_g(r, gr, dgr=dgr, **kwargs)
        q_ft, fq_ft, q, fq, r, gr, dfq_ft, dfq, dgr = self.g_using_F(
            r, gr, q, fq, cutoff, dgr=dgr, dfq=dfq, **kwargs)
        gr, dgr = self.converter.g_to_GK(r, gr, dgr=dgr, **kwargs)
        return q_ft, fq_ft, q, fq, r, gr, dfq_ft, dfq, dgr

    def GK_using_S(self, r, gr, q, sq, cutoff, dgr=None, dsq=None, **kwargs):
        """
        Fourier filters real space :math:`G_{Keen Version}(r)`
        using the reciprocal space :math:`S(Q)`

        :param r: :math:`r`-space vector
        :type r: numpy.array or list
        :param gr: :math:`G_{Keen Version}(r)` vector
        :type gr: numpy.array or list
        :param q: :math:`Q`-space vector
        :type q: numpy.array or list
        :param fq: :math:`S(Q)` vector
        :type fq: numpy.array or list
        :param cutoff: The :math:`r_{max}` value to filter from 0. to cutoff
        :type cutoff: float

        :return: A tuple of the :math:`Q` and :math:`S(Q)`
                 for the 0. to cutoff transform,
                 the corrected :math:`Q` and :math:`S(Q)`,
                 and the filtered :math:`r` and :math:`G_{Keen Version}(r)`.

                 Thus,
                 [
                    :math:`Q_{FF}`,
                    :math:`S(Q)_{FF}`,
                    :math:`Q`,
                    :math:`S(Q)`,
                    :math:`r_{FF}`,
                    :math:`G_{Keen Version}(r)_{FF}`
                ]
        :rtype: tuple of numpy.array
        """

        fq, dfq = self.converter.S_to_F(q, sq, dsq=dsq)
        q_ft, fq_ft, q, fq, r, gr, dfq_ft, dfq, dgr = self.GK_using_F(
            r, gr, q, fq, cutoff, dgr=dgr, dfq=dfq, **kwargs)
        sq_ft, dsq_ft = self.converter.F_to_S(q_ft, fq_ft, dfq=dfq_ft)
        sq, dsq = self.converter.F_to_S(q, fq, dfq=dfq)
        return q_ft, sq_ft, q, sq, r, gr, dsq_ft, dsq, dgr

    def GK_using_FK(self, r, gr, q, fq, cutoff, dgr=None, dfq=None, **kwargs):
        """
        Fourier filters real space :math:`G_{Keen Version}(r)`
        using the reciprocal space :math:`F(Q)`

        :param r: :math:`r`-space vector
        :type r: numpy.array or list
        :param gr: :math:`G_{Keen Version}(r)` vector
        :type gr: numpy.array or list
        :param q: :math:`Q`-space vector
        :type q: numpy.array or list
        :param fq: :math:`F(Q)` vector
        :type fq: numpy.array or list
        :param cutoff: The :math:`r_{max}` value to filter from 0. to cutoff
        :type cutoff: float

        :return: A tuple of the :math:`Q` and :math:`F(Q)`
                 for the 0. to cutoff transform,
                 the corrected :math:`Q` and :math:`F(Q)`,
                 and the filtered :math:`r` and :math:`G_{Keen Version}(r)`.

                 Thus,
                 [
                    :math:`Q_{FF}`,
                    :math:`F(Q)_{FF}`,
                    :math:`Q`,
                    :math:`F(Q)`,
                    :math:`r_{FF}`,
                    :math:`G_{Keen Version}(r)_{FF}`
                ]
        :rtype: tuple of numpy.array
        """

        fq, dfq = self.converter.FK_to_F(q, fq, dfq=dfq, **kwargs)
        q_ft, fq_ft, q, fq, r, gr, dfq_ft, dfq, dgr = self.GK_using_F(
            r, gr, q, fq, cutoff, dgr=dgr, dfq=dfq, **kwargs)
        fq_ft, dfq_ft = self.converter.F_to_FK(
            q_ft, fq_ft, dfq=dfq_ft, **kwargs)
        fq, dfq = self.converter.F_to_FK(q, fq, dfq=dfq, **kwargs)
        return q_ft, fq_ft, q, fq, r, gr, dfq_ft, dfq, dgr

    def GK_using_DCS(self, r, gr, q, dcs, cutoff, dgr=None, ddcs=None,
                     **kwargs):
        """
        Fourier filters real space :math:`G_{Keen Version}(r)`
        using the reciprocal space
        :math:`\\frac{d \\sigma}{d \\Omega}(Q)`

        :param r: :math:`r`-space vector
        :type r: numpy.array or list
        :param gr: :math:`G_{Keen Version}(r)` vector
        :type gr: numpy.array or list
        :param q: :math:`Q`-space vector
        :type q: numpy.array or list
        :param dcs: :math:`\\frac{d \\sigma}{d \\Omega}(Q)` vector
        :type dcs: numpy.array or list
        :param cutoff: The :math:`r_{max}` value to filter from 0. to cutoff
        :type cutoff: float

        :return: A tuple of the :math:`Q` and
                 :math:`\\frac{d \\sigma}{d \\Omega}(Q)`
                 for the 0. to cutoff transform,
                 the corrected :math:`Q` and
                 :math:`\\frac{d \\sigma}{d \\Omega}(Q)`,
                 and the filtered :math:`r` and :math:`G_{Keen Version}(r)`.

                 Thus,
                 [:math:`Q_{FF}`, :math:`\\frac{d \\sigma}{d \\Omega}(Q)_{FF}`,
                 :math:`Q`, :math:`\\frac{d \\sigma}{d \\Omega}(Q)`,
                 :math:`r_{FF}`, :math:`G_{Keen Version}(r)_{FF}]`
        :rtype: tuple of numpy.array
        """
        fq, dfq = self.converter.DCS_to_F(q, dcs, ddcs=ddcs, **kwargs)
        q_ft, fq_ft, q, fq, r, gr, dfq_ft, dfq, dgr = self.GK_using_F(
            r, gr, q, fq, cutoff, dgr=dgr, dfq=dfq, **kwargs)
        dcs_ft, ddcs_ft = self.converter.F_to_DCS(
            q_ft, fq_ft, dfq=dfq_ft, **kwargs)
        dcs, ddcs = self.converter.F_to_DCS(q, fq, dfq=dfq, **kwargs)
        return q_ft, dcs_ft, q, dcs, r, gr, ddcs_ft, ddcs, dgr



import unittest
import numpy
from numpy.testing import assert_array_equal, assert_allclose
from tests.utils import \
    load_data, get_index_of_function
from tests.materials import Nickel, Argon
from pystog.utils import \
    RealSpaceHeaders, ReciprocalSpaceHeaders
from pystog.transformer import Transformer

# Real Space Function


class TestTransformerBase(unittest.TestCase):
    rtol = 1e-2
    atol = 1e-2

    def initialize_material(self):
        # setup input data
        self.kwargs = self.material.kwargs

        # setup the first, last indices
        self.real_space_first = self.material.real_space_first
        self.real_space_last = self.material.real_space_last

        data = load_data(self.material.real_space_filename)
        self.r = data[:, get_index_of_function("r", RealSpaceHeaders)]
        self.gofr = data[:, get_index_of_function("g(r)", RealSpaceHeaders)]
        self.GofR = data[:, get_index_of_function("G(r)", RealSpaceHeaders)]
        self.GKofR = data[:, get_index_of_function("GK(r)", RealSpaceHeaders)]

        # targets for 1st peaks
        self.gofr_target = self.material.gofr_target
        self.GofR_target = self.material.GofR_target
        self.GKofR_target = self.material.GKofR_target

        # setup the tolerance
        self.reciprocal_space_first = self.material.reciprocal_space_first
        self.reciprocal_space_last = self.material.reciprocal_space_last

        data = load_data(self.material.reciprocal_space_filename)
        self.q = data[:, get_index_of_function("Q", ReciprocalSpaceHeaders)]
        self.sq = data[:, get_index_of_function(
            "S(Q)", ReciprocalSpaceHeaders)]
        self.fq = data[:, get_index_of_function(
            "Q[S(Q)-1]", ReciprocalSpaceHeaders)]
        self.fq_keen = data[:, get_index_of_function(
            "FK(Q)", ReciprocalSpaceHeaders)]
        self.dcs = data[:, get_index_of_function(
            "DCS(Q)", ReciprocalSpaceHeaders)]

        # targets for 1st peaks
        self.sq_target = self.material.sq_target
        self.fq_target = self.material.fq_target
        self.fq_keen_target = self.material.fq_keen_target
        self.dcs_target = self.material.dcs_target

    def setUp(self):
        unittest.TestCase.setUp(self)
        self.transformer = Transformer()

        # Parameters for fourier transform testing
        self._ft_first = 28
        self._ft_last = 35
        fs = 100  # sample rate
        f = 10  # the frequency of the signal
        self._ft_xin = numpy.linspace(0.0, 100., 1000)
        self._ft_yin = numpy.asarray(
            [numpy.sin(2 * numpy.pi * f * (i / fs)) for i in self._ft_xin])
        self._ft_xout = numpy.linspace(0.0, 2.0, 100)

    def tearDown(self):
        unittest.TestCase.tearDown(self)

    # Utilities

    def test_apply_cropping(self):
        xin = numpy.linspace(0.5, 1.0, 11)
        yin = numpy.linspace(4.5, 5.0, 11)
        x, y, _ = self.transformer.apply_cropping(xin, yin, 0.6, 0.7)
        assert_array_equal(x, [0.6, 0.65, 0.7])
        assert_array_equal(y, [4.6, 4.65, 4.7])

    def test_fourier_transform(self):
        xout, yout, _ = self.transformer.fourier_transform(self._ft_xin,
                                                           self._ft_yin,
                                                           self._ft_xout)
        yout_target = [-0.14265772,
                       -10.8854444,
                       18.13582784,
                       49.72976782,
                       26.3590524,
                       -8.08540764,
                       -3.38810001]
        assert_allclose(yout[self._ft_first:self._ft_last],
                        yout_target,
                        rtol=self.rtol, atol=self.atol)

    def test_fourier_transform_with_lorch(self):
        kwargs = {"lorch": True}
        xout, yout, _ = self.transformer.fourier_transform(self._ft_xin,
                                                           self._ft_yin,
                                                           self._ft_xout,
                                                           **kwargs)
        yout_target = [-1.406162,
                       3.695632,
                       18.788041,
                       29.370677,
                       21.980533,
                       6.184271,
                       -1.234159]
        assert_allclose(yout[self._ft_first:self._ft_last],
                        yout_target,
                        rtol=self.rtol, atol=self.atol)

    def test_fourier_transform_with_low_x(self):
        kwargs = {"OmittedXrangeCorrection": True}
        xout, yout, _ = self.transformer.fourier_transform(self._ft_xin,
                                                           self._ft_yin,
                                                           self._ft_xout,
                                                           **kwargs)
        yout_target = [-0.142658,
                       -10.885444,
                       18.135828,
                       49.729768,
                       26.359052,
                       -8.085408,
                       -3.388100]
        assert_allclose(yout[self._ft_first:self._ft_last],
                        yout_target,
                        rtol=self.rtol, atol=self.atol)

    def test_low_x_correction(self):
        kwargs = {"lorch": False}
        xout, yout, _ = self.transformer.fourier_transform(self._ft_xin,
                                                           self._ft_yin,
                                                           self._ft_xout,
                                                           **kwargs)
        yout = self.transformer._low_x_correction(self._ft_xin,
                                                  self._ft_yin,
                                                  xout, yout,
                                                  **kwargs)
        yout_target = [-0.142658,
                       -10.885444,
                       18.135828,
                       49.729768,
                       26.359052,
                       -8.085408,
                       -3.388100]
        assert_allclose(yout[self._ft_first:self._ft_last],
                        yout_target,
                        rtol=self.rtol, atol=self.atol)

    def test_low_x_correction_with_lorch(self):
        kwargs = {"lorch": True}
        xout, yout, _ = self.transformer.fourier_transform(self._ft_xin,
                                                           self._ft_yin,
                                                           self._ft_xout,
                                                           **kwargs)
        yout = self.transformer._low_x_correction(self._ft_xin,
                                                  self._ft_yin,
                                                  xout, yout,
                                                  **kwargs)
        yout_target = [-1.406162,
                       3.695632,
                       18.788041,
                       29.370677,
                       21.980533,
                       6.184271,
                       -1.234159]
        assert_allclose(yout[self._ft_first:self._ft_last],
                        yout_target,
                        rtol=self.rtol, atol=self.atol)

    # Real space

    # g(r) tests

    def g_to_S(self):
        q, sq, _ = self.transformer.g_to_S(
            self.r, self.gofr, self.q, **self.kwargs)
        first, last = self.reciprocal_space_first, self.reciprocal_space_last
        assert_allclose(sq[first:last],
                        self.sq_target,
                        rtol=self.rtol, atol=self.atol)

    def g_to_F(self):
        q, fq, _ = self.transformer.g_to_F(
            self.r, self.gofr, self.q, **self.kwargs)
        first, last = self.reciprocal_space_first, self.reciprocal_space_last
        assert_allclose(fq[first:last],
                        self.fq_target,
                        rtol=self.rtol, atol=self.atol)

    def g_to_FK(self):
        q, fq_keen, _ = self.transformer.g_to_FK(
            self.r, self.gofr, self.q, **self.kwargs)
        first, last = self.reciprocal_space_first, self.reciprocal_space_last
        assert_allclose(fq_keen[first:last],
                        self.fq_keen_target,
                        rtol=self.rtol, atol=self.atol)

    def g_to_DCS(self):
        q, dcs, _ = self.transformer.g_to_DCS(
            self.r, self.gofr, self.q, **self.kwargs)
        first, last = self.reciprocal_space_first, self.reciprocal_space_last
        assert_allclose(dcs[first:last],
                        self.dcs_target,
                        rtol=self.rtol, atol=self.atol)

    # G(r) tests

    def G_to_S(self):
        q, sq, _ = self.transformer.G_to_S(
            self.r, self.GofR, self.q, **self.kwargs)
        first, last = self.reciprocal_space_first, self.reciprocal_space_last
        assert_allclose(sq[first:last],
                        self.sq_target,
                        rtol=self.rtol, atol=self.atol)

    def G_to_F(self):
        q, fq, _ = self.transformer.G_to_F(
            self.r, self.GofR, self.q, **self.kwargs)
        first, last = self.reciprocal_space_first, self.reciprocal_space_last
        assert_allclose(fq[first:last],
                        self.fq_target,
                        rtol=self.rtol, atol=self.atol)

    def G_to_FK(self):
        q, fq_keen, _ = self.transformer.G_to_FK(
            self.r, self.GofR, self.q, **self.kwargs)
        first, last = self.reciprocal_space_first, self.reciprocal_space_last
        assert_allclose(fq_keen[first:last],
                        self.fq_keen_target,
                        rtol=self.rtol, atol=self.atol)

    def G_to_DCS(self):
        q, dcs, _ = self.transformer.G_to_DCS(
            self.r, self.GofR, self.q, **self.kwargs)
        first, last = self.reciprocal_space_first, self.reciprocal_space_last
        assert_allclose(dcs[first:last],
                        self.dcs_target,
                        rtol=self.rtol, atol=self.atol)
    # GK(r) tests

    def GK_to_S(self):
        q, sq, _ = self.transformer.GK_to_S(
            self.r, self.GKofR, self.q, **self.kwargs)
        first, last = self.reciprocal_space_first, self.reciprocal_space_last
        assert_allclose(sq[first:last],
                        self.sq_target,
                        rtol=self.rtol, atol=self.atol)

    def GK_to_F(self):
        q, fq, _ = self.transformer.GK_to_F(
            self.r, self.GKofR, self.q, **self.kwargs)
        first, last = self.reciprocal_space_first, self.reciprocal_space_last
        assert_allclose(fq[first:last],
                        self.fq_target,
                        rtol=self.rtol, atol=self.atol)

    def GK_to_FK(self):
        q, fq_keen, _ = self.transformer.GK_to_FK(
            self.r, self.GKofR, self.q, **self.kwargs)
        first, last = self.reciprocal_space_first, self.reciprocal_space_last
        assert_allclose(fq_keen[first:last],
                        self.fq_keen_target,
                        rtol=self.rtol, atol=self.atol)

    def GK_to_DCS(self):
        q, dcs, _ = self.transformer.GK_to_DCS(
            self.r, self.GKofR, self.q, **self.kwargs)
        first, last = self.reciprocal_space_first, self.reciprocal_space_last
        assert_allclose(dcs[first:last],
                        self.dcs_target,
                        rtol=self.rtol, atol=self.atol)

    # Reciprocal space

    # S(Q) tests
    def S_to_g(self):
        r, gofr, _ = self.transformer.S_to_g(
            self.q, self.sq, self.r, **self.kwargs)
        first, last = self.real_space_first, self.real_space_last
        assert_allclose(gofr[first:last],
                        self.gofr_target,
                        rtol=self.rtol, atol=self.atol)

    def S_to_G(self):
        r, GofR, _ = self.transformer.S_to_G(
            self.q, self.sq, self.r, **self.kwargs)
        first, last = self.real_space_first, self.real_space_last
        assert_allclose(GofR[first:last],
                        self.GofR_target,
                        rtol=self.rtol, atol=self.atol)

    def S_to_GK(self):
        r, GKofR, _ = self.transformer.S_to_GK(
            self.q, self.sq, self.r, **self.kwargs)
        first, last = self.real_space_first, self.real_space_last
        assert_allclose(GKofR[first:last],
                        self.GKofR_target,
                        rtol=self.rtol, atol=self.atol)
    # Q[S(Q)-1] tests

    def F_to_g(self):
        r, gofr, _ = self.transformer.F_to_g(
            self.q, self.fq, self.r, **self.kwargs)
        first, last = self.real_space_first, self.real_space_last
        assert_allclose(gofr[first:last],
                        self.gofr_target,
                        rtol=self.rtol, atol=self.atol)

    def F_to_G(self):
        r, GofR, _ = self.transformer.F_to_G(
            self.q, self.fq, self.r, **self.kwargs)
        first, last = self.real_space_first, self.real_space_last
        assert_allclose(GofR[first:last],
                        self.GofR_target,
                        rtol=self.rtol, atol=self.atol)

    def F_to_GK(self):
        r, GKofR, _ = self.transformer.F_to_GK(
            self.q, self.fq, self.r, **self.kwargs)
        first, last = self.real_space_first, self.real_space_last
        assert_allclose(GKofR[first:last],
                        self.GKofR_target,
                        rtol=self.rtol, atol=self.atol)
    # FK(Q) tests

    def FK_to_g(self):
        r, gofr, _ = self.transformer.FK_to_g(
            self.q, self.fq_keen, self.r, **self.kwargs)
        first, last = self.real_space_first, self.real_space_last
        assert_allclose(gofr[first:last],
                        self.gofr_target,
                        rtol=self.rtol, atol=self.atol)

    def FK_to_G(self):
        r, GofR, _ = self.transformer.FK_to_G(
            self.q, self.fq_keen, self.r, **self.kwargs)
        first, last = self.real_space_first, self.real_space_last
        assert_allclose(GofR[first:last],
                        self.GofR_target,
                        rtol=self.rtol, atol=self.atol)

    def FK_to_GK(self):
        r, GKofR, _ = self.transformer.FK_to_GK(
            self.q, self.fq_keen, self.r, **self.kwargs)
        first, last = self.real_space_first, self.real_space_last
        assert_allclose(GKofR[first:last],
                        self.GKofR_target,
                        rtol=self.rtol, atol=self.atol)
    # DCS(Q) tests

    def DCS_to_g(self):
        r, gofr, _ = self.transformer.DCS_to_g(
            self.q, self.dcs, self.r, **self.kwargs)
        first, last = self.real_space_first, self.real_space_last
        assert_allclose(gofr[first:last],
                        self.gofr_target,
                        rtol=self.rtol, atol=self.atol)

    def DCS_to_G(self):
        r, GofR, _ = self.transformer.DCS_to_G(
            self.q, self.dcs, self.r, **self.kwargs)
        first, last = self.real_space_first, self.real_space_last
        assert_allclose(GofR[first:last],
                        self.GofR_target,
                        rtol=self.rtol, atol=self.atol)

    def DCS_to_GK(self):
        r, GKofR, _ = self.transformer.DCS_to_GK(
            self.q, self.dcs, self.r, **self.kwargs)
        first, last = self.real_space_first, self.real_space_last
        assert_allclose(GKofR[first:last],
                        self.GKofR_target,
                        rtol=self.rtol, atol=self.atol)


class TestTransformerNickel(TestTransformerBase):
    def setUp(self):
        super(TestTransformerNickel, self).setUp()
        self.material = Nickel()
        self.initialize_material()

    def test_g_to_S(self):
        self.g_to_S()

    def test_g_to_F(self):
        self.g_to_F()

    def test_g_to_FK(self):
        self.g_to_FK()

    def test_g_to_DCS(self):
        self.g_to_DCS()

    def test_G_to_S(self):
        self.G_to_S()

    def test_G_to_F(self):
        self.G_to_F()

    def test_G_to_FK(self):
        self.G_to_FK()

    def test_G_to_DCS(self):
        self.G_to_DCS()

    def test_GK_to_S(self):
        self.GK_to_S()

    def test_GK_to_F(self):
        self.GK_to_F()

    def test_GK_to_FK(self):
        self.GK_to_FK()

    def test_GK_to_DCS(self):
        self.GK_to_DCS()


class TestTransformerArgon(TestTransformerBase):
    def setUp(self):
        super(TestTransformerArgon, self).setUp()
        self.material = Argon()
        self.initialize_material()

    def test_g_to_S(self):
        self.g_to_S()

    def test_g_to_F(self):
        self.g_to_F()

    def test_g_to_FK(self):
        self.g_to_FK()

    def test_g_to_DCS(self):
        self.g_to_DCS()

    def test_G_to_S(self):
        self.G_to_S()

    def test_G_to_F(self):
        self.G_to_F()

    def test_G_to_FK(self):
        self.G_to_FK()

    def test_G_to_DCS(self):
        self.G_to_DCS()

    def test_GK_to_S(self):
        self.GK_to_S()

    def test_GK_to_F(self):
        self.GK_to_F()

    def test_GK_to_FK(self):
        self.GK_to_FK()

    def test_GK_to_DCS(self):
        self.GK_to_DCS()

    def test_S_to_g(self):
        self.S_to_g()

    def test_S_to_G(self):
        self.S_to_G()

    def test_S_to_GK(self):
        self.S_to_GK()

    def test_F_to_g(self):
        self.F_to_g()

    def test_F_to_G(self):
        self.F_to_G()

    def test_F_to_GK(self):
        self.F_to_GK()

    def test_FK_to_g(self):
        self.FK_to_g()

    def test_FK_to_G(self):
        self.FK_to_G()

    def test_FK_to_GK(self):
        self.FK_to_GK()

    def test_DCS_to_g(self):
        self.DCS_to_g()

    def test_DCS_to_G(self):
        self.DCS_to_G()

    def test_DCS_to_GK(self):
        self.DCS_to_GK()


if __name__ == '__main__':
    unittest.main()  # pragma: no cover
